from GangaCore.Core.exceptions import SplitterError
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Lib.Job import Job
from .SplitterUtils import DatasetSplitter
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
#import copy
import GangaCore.Utility.logging
# from GangaGaudi.Lib.Datasets import GaudiDataset

logger = GangaCore.Utility.logging.getLogger()


class GaudiInputDataSplitter(ISplitter):

    """Splits a job into sub-jobs by partitioning the input data

    GaudiInputDataSplitter can be used to split a job into multiple subjobs, where
    each subjob gets an unique subset of the inputdata files.
    """
    _name = 'GaudiInputDataSplitter'
    _schema = Schema(Version(1, 0), {
        'filesPerJob': SimpleItem(defvalue=10,
                                  doc='Number of files per subjob',
                                  typelist=['int']),
        'maxFiles': SimpleItem(defvalue=None,
                               doc='Maximum number of files to use in a masterjob (None = all files)',
                               typelist=['int', 'type(None)'])
    })

    # returns splitter generator can be overridden for modified behaviour
    # job is passed in only so that it has full information at the time of splitting
    # this may be useful if generator must change dependent on job properties like
    # backend for example. (see GangaLHCb.Lib.Splitters.SplitByFiles.py)
    def _splitter(self, job, inputdata):
        if (inputdata is None) or (len(inputdata.files) == 0):
            logger.error('Cannot split if no inputdata given!')
            raise SplitterError('inputdata is None')
        logger.debug("Found %s files inputdata" % str(len(inputdata.files)))
        return DatasetSplitter(inputdata, self.filesPerJob, self.maxFiles)

    # returns a subjob based on the master and the reduced dataset can be overridden
    # for modified behaviour
    def _create_subjob(self, job, dataset):
        logger.debug("_create_subjob")
        j = Job()
        j.copyFrom(job)
        j.splitter = None
        j.merger = None
        j.inputsandbox = []  # master added automatically
        j.inputdata = GaudiDataset(files=dataset)
##         if not j.inputdata: j.inputdata = GaudiDataset(files=dataset)
# else:               j.inputdata.files = dataset
        return j

    def split(self, job):
        logger.debug("split")
        if self.filesPerJob < 1:
            logger.error('filesPerJob must be greater than 0.')
            raise SplitterError('filesPerJob < 1 : %d' % self.filesPerJob)
        elif (self.maxFiles is not None) and self.maxFiles < 1:
            logger.error('maxFiles must be greater than 0.')
            raise SplitterError('maxFiles < 1 : %d' % self.maxFiles)

        subjobs = []

        logger.debug("Creating all_jobs")
        all_jobs = self._splitter(job, job.inputdata)

        logger.info("Constructing subjobs")
        logger.debug("Filling DataSet")
        for dataset in all_jobs:
            logger.debug("Creating Subjobs with dataset of size: %s" % str(len(dataset)))
            #logger.debug( "Creating Subjobs with dataset: %s" % str(dataset) )
            subjobs.append(self._create_subjob(job, dataset))

        logger.info("Finished Splitting")
        logger.debug("Returning all subjobs")
        return subjobs



from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.Tasks.common import *
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
import time
from GangaCore.GPIDev.Lib.Tasks import ITask

########################################################################                                                                                                                                                                     

class ND280Task(ITask):
    """T2K add-ons for the Task framework"""
    _schema = Schema(Version(1,0), dict(ITask._schema.datadict.items()))

    _category = 'tasks'
    _name = 'ND280Task'
    _exportmethods = ITask._exportmethods + [ ]

    _tasktype = "ITask"

    default_registry = "tasks"



from GangaCore.Utility.logging import getLogger
l = getLogger()
import time


def f(name=''):
    for i in range(5):
        l.warning(name + str(i))
        time.sleep(1)

f('Main:')

from GangaCore.Core.GangaThread import GangaThread


class MyThread(GangaThread):

    def run(self):
        f('GangaThread:')

t = MyThread('GangaThread')
t.start()




import os

from GangaCore.Runtime.GPIexport import exportToGPI
from GangaCore.GPIDev.Base.Proxy import GPIProxyClassFactory

#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/
from GangaLHCb.Lib.Applications import AppsBaseUtils
from GangaCore.Utility.logging import getLogger

from .GaudiPython import GaudiPython

from .Bender import Bender
from .BenderScript import BenderScript
from .Ostap import Ostap

from .BenderBox import BenderModule, BenderRun, OstapRun
from .GaudiExec import GaudiExec

# Add any additional Packages required by the user in the .gangarc file
from GangaCore.Utility.Config import getConfig

logger = getLogger()
logger.debug("User Added Apps")

config = getConfig('LHCb')
user_added = config['UserAddedApplications']
user_apps = user_added.split(':')
if user_apps == user_added and len(user_added) > 0:
    AppsBaseUtils.addNewLHCbapp(user_apps)
for app in user_apps:
    if len(app) > 0:
        AppsBaseUtils.addNewLHCbapp(app)

logger.debug("Constructing AppsBase Apps")
f = open(os.path.join(os.path.dirname(__file__), 'AppsBase.py'), 'r')
cls = f.read()
f.close()
all_apps = ''
for app in AppsBaseUtils.available_apps():
    if app in dir():
        continue
    app = str(app)
    this_app = cls.replace('AppName', app)
    all_apps = all_apps + str('\n\n') + this_app
    #exec(this_exec, all_global, all_local)
    logger.debug("Adding %s" % str(app))

logger.debug("Adding apps")
modules= compile(all_apps, '<string>', 'exec')
exec(modules)

logger.debug("Fin")





##########################################################################
# Ganga Project. http://cern.ch/ganga
#
##########################################################################

from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException
from GangaCore.GPIDev.Adapters.IChecker import IFileChecker
from GangaCore.GPIDev.Schema import SimpleItem
from GangaCore.Utility.logging import getLogger
import subprocess
import copy
import os
import re


logger = getLogger()


class FileChecker(IFileChecker):

    """
    Checks if string is in file.
    self.searchStrings are the files you would like to check for.
    self.files are the files you would like to check.
    self.failIfFound (default = True) decides whether to fail the job if the string is found. If you set this to false the job will fail if the string *isnt* found.
    self.fileMustExist toggles whether to fail the job if the specified file doesn't exist (default is True).
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['searchStrings'] = SimpleItem(
        defvalue=[], doc='String to search for')
    _schema.datadict['failIfFound'] = SimpleItem(
        True, doc='Toggle whether job fails if string is found or not found.')
    _category = 'postprocessor'
    _name = 'FileChecker'
    _exportmethods = ['check']

    def check(self, job):
        """
        Check that a string is in a file, takes the job object as input.
        """

        if not len(self.searchStrings):
            raise PostProcessException('No searchStrings specified, FileChecker will do nothing!')
        filepaths = self.findFiles(job)
        if not len(filepaths):
            raise PostProcessException('None of the files to check exist, FileChecker will do nothing!')
        for filepath in filepaths:
            for searchString in self.searchStrings:
                stringFound = False
                # self.findFiles() guarantees that file at filepath exists,
                # hence no exception handling
                with open(filepath) as file:
                    for line in file:
                        if re.search(searchString, line):
                            if self.failIfFound is True:
                                logger.info(
                                    'The string %s has been found in file %s, FileChecker will fail job(%s)', searchString, filepath, job.fqid)
                                return self.failure
                            stringFound = True
                if not stringFound and self.failIfFound is False:
                    logger.info('The string %s has not been found in file %s, FileChecker will fail job(%s)', searchString, filepath, job.fqid)
                    return self.failure
        return self.result





# Tools to ship with every rat job.
# Useful!

import os
import sys
import re
import subprocess


###########################################
# Exceptions
###########################################
class JobToolsException(Exception):
    def __init__(self, error):
        Exception.__init__(self, error)


class CommandException(JobToolsException):
    def __init__(self, command, out, err):
        '''Pass in the command failed and outputs/errors
        '''
        error = '\nCommand: %s' % command
        error += '\nOutput: %s' % out
        error += '\nError: %s' % err
        JobToolsException.__init__(self, error)


###########################################
# Other classes
###########################################
class JobHelper(object):
    '''A singleton class to persist information relating to job setup.

    Useful for jobs where multiple functions need to be called that all
    need the same environment!
    '''
    
    _instance = None

    class SingletonHelper:
        def __call__(self, *args, **kw):
            if JobHelper._instance is None:
                object = JobHelper()
                JobHelper._instance = object
            return JobHelper._instance

    get_instance = SingletonHelper()

    def __init__(self):
        '''Nothing done in the constructor'''
        self._envs = []
        self._vars = {}

    def add_environment(self, path):
        '''Append an environment file that should be used.

        These must be added in the correct order.'''
        if not os.path.exists(path):
            raise JobToolsException("JobHelper: no such file %s" % path)
        self._envs.append(path)

    def set_variables(self, **kwargs):
        '''Add an environment variable; these are set after before environment files are sourced.
        '''
        for key in kwargs:
            self._vars[key] = kwargs[key]

    def get_environment_script(self):
        '''Return the environment as required for jobs'''
        command = ""
        for env in self._envs:
            command += "source %s\n" % env
        return command

    def get_environment_variables(self):
        '''Return the environment variables as required for jobs
        '''
        return self._vars


###########################################
# Execution commands
###########################################
def execute_ratenv(command, args, cwd=None, exit_if_fail=True):
    '''Sources appropriate environment information prior to call.
    
    Provide string command and list of arguments as with "execute".
    This is the only command that raises an exception if if fails by default.
    '''
    job_helper = JobHelper.get_instance()
    script = job_helper.get_environment_script()
    script += command
    for arg in args:
        script += ' %s' % arg
    script += '\n'
    return execute_complex(script, cwd=cwd, exit_if_fail=exit_if_fail)


def execute_complex(script, cwd=None, exit_if_fail=False, login_script=False):
    '''Write a temporary script, run, remove.
    '''
    if cwd is None:
        cwd = os.getcwd()
    filename = os.path.join(cwd, "temp.sh")
    temp_file = file(filename, 'w')
    temp_file.write(script)
    temp_file.close()
    # Don't pass the exit_if_fail on to execute, rather print out info here
    # Otherwise just get confusing errors regarding temp.sh
    args = [filename]
    if login_script:
        args = ['-l', filename]
    rtc, out, err = execute('/bin/bash', args)
    if exit_if_fail is True and rtc!=0:
        raise CommandException(script, out, err)
    return rtc, out, err


def execute(command, args, env=None, cwd=None, exit_if_fail=False):
    '''Simple command.
    '''
    shell_command = [command] + args
    # Default to current environment, then apply overrides from job_helper
    # then apply overrides from function arguments
    use_env = os.environ
    job_helper = JobHelper.get_instance()
    for env in job_helper.get_environment_variables():
        use_env[key] = env[key]
    if env is not None:
        for key in env:
            use_env[key] = env[key]
    if cwd is None:
        cwd = os.getcwd()
    for i,arg in enumerate(args):
        if type(arg) is unicode:
            args[i] = ucToStr(arg) # urgh
    process = subprocess.Popen(args=shell_command, env=use_env, cwd=cwd, stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=False)
    output, error = process.communicate()
    if process.returncode != 0:
        sys.stderr.write("Process failed: %s \n" % shell_command)
        sys.stderr.write("Output: %s \n" % output)
        sys.stderr.write("Error: %s \n" % error)
        if exit_if_fail is True:
            raise CommandException(shell_command, output, error)
    output = output.split('\n') # want lists
    error = error.split('\n') # will always have a '' as last element, unless :
    if output[len(output)-1]=='':
        del output[-1]
    if error[len(error)-1]=='':
        del error[-1]
    return process.returncode, output, error # don't always fail on returncode!=0


###########################################
# Data transfer and checking tools
###########################################
def copy_data(file_list, output_dir, grid_mode):
    '''Transfer files from file list to the given directory.
    
    TODO: exit if upload fails, but first delete any files that have already been uploaded.
    '''
    dump_out = {}
    if grid_mode is not None:

        if grid_mode=='lcg':            
            lfc_dir = os.path.join('lfn:/grid/snoplus.snolab.ca', output_dir)
            execute('lfc-mkdir', [lfc_dir.lstrip('lfn:')], exit_if_fail=False)
            se_name = os.environ['VO_SNOPLUS_SNOLAB_CA_DEFAULT_SE']
            args = ['--list-se', '--vo', 'snoplus.snolab.ca', '--attrs', 'VOInfoPath', '--query', 'SE=%s' % (se_name)]
            rtc, out, err = execute('lcg-info', args)
            srm_path = "INCOMPLETE" # in case we aren't able to get the path
            if rtc==0:
                # remove the leading / so that os.path.join works
                srm_path = out[-2].split()[-1].lstrip('/')
            for filename in file_list:
                dump_out[filename] = {}
                command = 'lcg-cr'
                se_path = os.path.join(output_dir, filename)
                lfc_path= os.path.join(lfc_dir, filename)
                args = ['--vo', 'snoplus.snolab.ca', '--checksum', '-d', se_name, '-P', se_path, '-l', lfc_path, filename]
                rtc, out, err = execute(command, args)
                dump_out[filename]['guid'] = out[0]
                dump_out[filename]['se'] = os.path.join('srm://%s' % se_name, srm_path, se_path)
                dump_out[filename]['size'] = os.stat(filename).st_size
                dump_out[filename]['cksum'] = adler32(filename)
                dump_out[filename]['name'] = se_name
                dump_out[filename]['lfc'] = lfc_path

        elif grid_mode=='srm':
            se_name = 'sehn02.atlas.ualberta.ca'
            surl = 'srm://sehn02.atlas.ualberta.ca/pnfs/atlas.ualberta.ca/data/snoplus'
            for filename in file_list:
                dump_out[filename] = {}
                command = 'lcg-cr'
                se_rel_path = os.path.join(output_dir, filename)
                se_full_path = os.path.join(surl, se_rel_path)
                lfc_dir = os.path.join('lfn:/grid/snoplus.snolab.ca', output_dir)
                lfc_path = os.path.join(lfc_dir, filename)
                args = ['--vo', 'snoplus.snolab.ca', '--checksum', '-d', se_full_path, '-l', lfc_path, filename]
                rtc,out,err = execute(command, args)
                dump_out[filename]['guid'] = out[0]
                dump_out[filename]['se'] = se_full_path
                dump_out[filename]['size'] = os.stat(filename).st_size
                dump_out[filename]['cksum'] = adler32(filename)
                dump_out[filename]['name'] = se_name
                dump_out[filename]['lfc'] = lfc_path
        else:
            raise Exception('Error, bad copy mode: %s' % grid_mode)
    else:
        import shutil
        import socket
        # Just run a local copy
        if not os.path.exists(output_dir):
            os.makedirs(output_dir)
        for filename in file_list:
            shutil.copy2(filename, output_dir)
            dump_out[filename] = {}
            dump_out[filename]['se'] = '%s:%s' % (socket.getfqdn(), os.path.join(output_dir, filename))
            dump_out[filename]['size'] = os.stat(filename).st_size
            dump_out[filename]['cksum'] = adler32(filename)
            dump_out[filename]['name'] = socket.getfqdn()
    return dump_out


def get_data(file_list, input_dir, grid_mode):
    '''Get files from a list and transfer to a local directory.

    TODO: ensure that this tries to access to closest local copy.
    '''
    if grid_mode is not None:
        # Both SRM and LCG mode should be fine using the LFC
        # TODO: ensure that the closest local copy is retrieved!
        lfc_dir = os.path.join('lfn:/grid/snoplus.snolab.ca', input_dir)
        for fin in file_list:
            fin = fin.strip() # Remove any spaces or other junk
            lfc_path = os.path.join(lfc_dir, fin)
            args = ['--vo','snoplus.snolab.ca', lfc_path, fin]
            execute('lcg-cp', args, exit_if_true=True)
    else:
        import shutil
        for fin in file_list:
            fin = fin.strip()#remove any spaces or other junk
            input_path = os.path.join(input_dir, fin)
            shutil.copy2(input_path, fin)


###########################################
# File utilities
###########################################
def adler32(filename):
    '''Calculate and return Alder32 checksum of file.
    '''
    import zlib
    block = 32 * 1024 * 1024
    val = 1
    f = open(filename, 'rb')
    while True:
        line = f.read(block)
        if len(line) == 0:
            break
        val = zlib.adler32(line, val)
        if val < 0:
            val += 2**32
    f.close()
    return hex(val)[2:10].zfill(8).lower()


def download_file(url, token=None, filename=None , cache_path=None):
    '''Download a file from a URL.  No user/password auth is allowed (security).  Tokens only.
    '''
    import urllib2
    if filename is None:
        filename = url.split('/')[-1]
    temp_file = os.path.join(cache_path, "download-temp")
    url_request = urllib2.Request(url)
    if token is not None:
        # Add OAuth authentication
        url_request.add_header("Authorization", "token %s" % token)
    remote_file = urllib2.urlopen(url_request) # Will raise exception if fails
    local_file = open(temp_file, 'wb')
    try:
        download_size = int(remote_file.info().getheaders("Content-Length")[0])
        downloaded = 0 # Amount downloaded
        block = 8192 # Convenient block size
        while True:
            buffer = remote_file.read(block)
            if not buffer: # Nothing left to download
                break
            downloaded += len(buffer)
            local_file.write(buffer)
        remote_file.close()
        local_file.close()
    except (KeyboardInterrupt, SystemExit):
        local_file.close()
        remote_file.close()
        os.remove(temp_file)
        raise
    if downloaded < download_size: # Something has gone wrong
        raise Exception("Download size error: %s <  %s" % (downloaded, download_size))
    os.rename(temp_file, os.path.join(cache_path, filename))


def untar_file(tarname, target_path, target_dir, cache_path, strip=0):
    '''Untar the file tarname to target_path, removing the first strip folders.
    '''
    import tarfile
    import shutil
    tarred_file = tarfile.open(os.path.join(cache_path, tarname))
    if strip==0:
        tarred_file.extractall(target_path)
        # Mote the file to the appropraite name and close the tar
        basedir = os.path.commonprefix(tarred_file.getnames())
        if basedir=='.':
            raise Exception("Cannot find a common base name")
        shutil.move(os.path.join(cache_path, basedir), os.path.join(cache_path, target_dir))
        tarred_file.close()
    else:
        # Move to temp dir, then target dir
        temp_dir = os.path.join(cache_path, 'temp_tar_dir')
        if os.path.exists(temp_dir):
            shutil.rmtree(temp_dir)
        tarred_file.extractall(temp_dir)
        tarred_file.close()
        # Now strip the dirs and move to the final location
        copy_dir = temp_dir
        for i_strip in range(0, strip):
            sub_folders = os.listdir(copy_dir)
            if 'pax_global_header' in sub_folders:
                sub_folders.remove( 'pax_global_header' )
            copy_dir = os.path.join(copy_dir, sub_folders)
        if os.path.exists(target_path):
            shutil.rmtree(target_path)
        shutil.copytree(copy_dir, target_path)
        shutil.rmtree(temp_dir)

###########################################
# Grid commands
###########################################
def exists(url):
    '''Just checks a file exists
    '''
    command = 'lcg-ls'
    rtc, out, err = execute(command, [url])
    if rtc!=0:
        raise JobToolsException('File does not exist %s: %s' % (url, err))


def lcg_ls(*args):
    '''List the contents of an lfn or SURL.
    This could be made a private function - users should use list_dir etc.
    '''
    command='lcg-ls'
    passargs = []
    for arg in args:
        passargs.append(arg)
    rtc,out,err = execute(command, passargs)
    return rtc, out, err


def getsize(url):
    '''Get the size (must be an srm)
    '''
    rtc, out, err = lcg_ls('-l', url)
    if rtc!=0:
        raise JobToolsException('File does not exist %s' % url)
    else:
        try:
            bits = out[0].split()
            size = int(bits[4])
            return size
        except:
            raise JobToolsException('Cannot get size of %s' % url)

        
def getguid(url):
    '''Get the guid of a file
    '''
    rtc,out,err = execute('lcg-lg', [url])
    if rtc!=0:
        raise JobToolsException('Cannot get url of %s'%url)
    else:
        guid = out[0]
        return guid


def listreps(url):
    '''Get the locations of a file in the lfc.
    Returns a list.
    '''
    rtc,out,err = execute('lcg-lr', [url])
    if rtc!=0:
        raise JobToolsException('lcg-lr error: %s'%url)
    else:
        return out


def checksum(url):
    '''Get the checksum of a file (adler32).
    Must provide an SRM.
    '''
    rtc,out,err = execute('lcg-get-checksum', [url])
    if rtc!=0:
        raise JobToolsException('Checksum error: %s'%url)
    else:
        try:
            cks = out[0].split()[0].strip()
            return cks
        except IndexError as e:
            raise JobToolsException('Checksum error: %s'%url)


def get_se_name(grid_mode):
    if grid_mode == 'srm':
        # Assume this means westgrid
        se_name = "sehn02.atlas.ualberta.ca"
    elif grid_mode == 'lcg':
        se_name = os.environ['VO_SNOPLUS_SNOLAB_CA_DEFAULT_SE']
    else:
        raise JobToolsException('Unknown grid mode %s; cannot get se name' % grid_mode)
    return se_name


def get_se_path(grid_mode):
    se_name = get_se_name(grid_mode)
    rtc, out, err = execute('lcg-info', ['--list-se', '--vo', 'snoplus.snolab.ca', '--attrs',
                                         'VOInfoPath', '--query', 'SE=%s' % (se_name)])
    if rtc!=0:
        raise JobToolsException('Cannot get se-path for %s' % (se_name))
    try:
        se_path = out[-2].split()[-1]
        if se_path[0]=='/':
            se_path = se_path[1:]
        return se_path
    except:
        raise JobToolsException('Cannot get se-path from %s' % (out))

###########################################
# Macro checks
###########################################
def check_processor(filename, processor):
    '''Check whether a processor is implemented.
    '''
    for line in open(filename, 'r').readlines():
        proc = '/proc/%s' % processor
        proc_last = '/proc_last/%s' % processor
        if proc in line or proc_last in line:
            if line.strip()[0]!='#':
                return True
    return False



################################################################################
# GangaND280 Project.
# Dima Vavilov
# Created 21/03/2016
################################################################################
"""@package ND280Control
Ganga module to execute runND280 from the nd280Control package.
"""

from GangaCore.GPIDev.Adapters.IPrepareApp import IPrepareApp
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Schema import *

from GangaCore.Utility.Config import getConfig

from GangaCore.GPIDev.Lib.File import *
from GangaCore.GPIDev.Lib.Registry.PrepRegistry import ShareRef
from GangaCore.GPIDev.Base.Proxy import isType
from GangaCore.Core.exceptions import ApplicationConfigurationError

import os, shutil, re
from GangaCore.Utility.files import expandfilename
shared_path = os.path.join(expandfilename(getConfig('Configuration')['gangadir']),'shared',getConfig('Configuration')['user'])

from . import ND280Configs

class runND280CosMC(IPrepareApp):
    """
    runND280CosMC application running runND280 from nd280Control with configuration files
    (created hereby) specific for Cosmic MC processing
        app = runND280CosMC()

    The required input for this module are:
        app.cmtsetup = '/home/me/myT2KWork/setup.sh'
        app.confopts = {'stage':'base', # stage can be 'base','fgd','tript' and 'all'
                        'nd280ver':'v11r31','num_events':'10000000'}

    NOTE: Ganga will run a bash script so the CMT setup script must be in sh/bash.

    It is also possible to pass command line arguments to the executable.
    For example to use '/home/me/tmp' as temporary directory:
        app.args = ['-t','/home/me/tmp']

    """
    _schema = Schema(Version(1,1), {
        'args' : SimpleItem(defvalue=[],typelist=['str','GangaCore.GPIDev.Lib.File.File.File','int'],sequence=1,strict_sequence=0,doc="List of arguments for the executable. Arguments may be strings, numerics or File objects."),
        'cmtsetup' : SimpleItem(defvalue=[],doc='Setup script(s) in bash to set up cmt and the cmt package of the executable.', typelist=['str'],sequence=1,strict_sequence=0),
        'confopts' : SimpleItem(defvalue={},doc='Options for configuration file', typelist=['str']),
        'env' : SimpleItem(defvalue={},typelist=['str'],doc='Environment'),
        'is_prepared' : SimpleItem(defvalue=None, strict_sequence=0, visitable=1, copyable=1, typelist=['type(None)','bool'],protected=0,comparable=1,doc='Location of shared resources. Presence of this attribute implies the application has been prepared.'),
        } )
    _category = 'applications'
    _name = 'runND280CosMC'
    _exportmethods = ['prepare']
    _GUIPrefs = [ { 'attribute' : 'args', 'widget' : 'String_List' },
                  { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                  { 'attribute' : 'confopts', 'widget' : 'String' },
                  { 'attribute' : 'env', 'widget' : 'DictOfString' } ]

    _GUIAdvancedPrefs = [ { 'attribute' : 'args', 'widget' : 'String_List' },
                          { 'attribute' : 'cmtsetup', 'widget' : 'String' },
                          { 'attribute' : 'confopts', 'widget' : 'String' },
                          { 'attribute' : 'env', 'widget' : 'DictOfString' } ]


    def __init__(self):
        super(runND280CosMC,self).__init__()


    def configure(self,masterappconfig):
        
        args = convertIntToStringArgs(self.args)

        job = self.getJobObject()

        if self.cmtsetup == []:
          raise ApplicationConfigurationError('No cmt setup script given.')

        for arg in args:
          if arg == '-c':
            raise ApplicationConfigurationError('Option "-c" given in args. You must use the configfile variable instead.')

        confopts = self.confopts


        # use input file from a "dataset"
        if job.inputdata is None:
            raise ApplicationConfigurationError('The given config file requires an input file but the inputdata of the job is not defined.')
        infiles = job.inputdata.get_dataset_filenames()
        if len(infiles) < 1:
            raise ApplicationConfigurationError('The given config file contains "inputfile" but not input file was given')
        if len(infiles) > 1:
            raise ApplicationConfigurationError('The given config file contains "inputfile" but more than one input file was given')
        confopts.update({'kinfile':infiles[0],'inputfile':infiles[0]})

        # extract "run number" from an input filename
        mtch = re.search(r"(\d{8})",os.path.basename(infiles[0]))
        if mtch:
            job.name = mtch.group(1)
            jobid    = mtch.group(1)
        else:
            raise  ApplicationConfigurationError('Can not extract run number')
        confopts.update({'run_number':str(int(jobid))})

        # create TND280Log config file (in job output dir)
        logConf  = "log.default.level = LogLevel\n"
        logConf += "error.default.level = SevereLevel\n"
        job.getOutputWorkspace().writefile(FileBuffer('nd280log.config',logConf),executable=0)

        # create config file
        cfg = ND280Configs.ND280Config('cosmicmc',confopts)
        inConf = cfg.CreateConfig()
        outConf = inConf
        job.getInputWorkspace().writefile(FileBuffer('nd280.cfg',outConf),executable=0)

        # create a script for a  backend
        args.append('-c')
        args.append(job.inputdir+'nd280.cfg')

        argsStr = ' '.join(args)
        # Create the bash script and put it in input dir.
        script = '#!/bin/bash\n'
        for f in self.cmtsetup:
            script += 'source '+f+'\n'
        script += 'cd '+job.outputdir+'\n'
        script += 'runND280 -t . '+argsStr+'\n'
        #script += 'echo runND280 '+argsStr+' > oa_cs_mu_00003333_' + confopts['stage'] + '.root\n'
        job.getInputWorkspace().writefile(FileBuffer('runND280.sh',script),executable=1)

        self._scriptname = job.inputdir+'runND280.sh'

        return (None,None)



config = getConfig('defaults_runND280') #_Properties
# config.options['exe'].type = type(None)


def convertIntToStringArgs(args):

    result = []
    
    for arg in args:
        if isinstance(arg,int):
            result.append(str(arg))
        else:
            result.append(arg)

    return result



class RTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.GPIDev.Adapters.StandardJobConfig import StandardJobConfig

        c = StandardJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)
        return c
        

class LCGRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.LCG import LCGJobConfig

        return LCGJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

class gLiteRTHandler(IRuntimeHandler):
    def prepare(self,app,appconfig,appmasterconfig,jobmasterconfig):
        from GangaCore.Lib.gLite import gLiteJobConfig

        return gLiteJobConfig(app._scriptname,app._getParent().inputsandbox,[],app._getParent().outputsandbox,app.env)

from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers

allHandlers.add('runND280CosMC','LSF', RTHandler)
allHandlers.add('runND280CosMC','Local', RTHandler)
allHandlers.add('runND280CosMC','PBS', RTHandler)
allHandlers.add('runND280CosMC','SGE', RTHandler)
allHandlers.add('runND280CosMC','Slurm', RTHandler)
allHandlers.add('runND280CosMC','Condor', RTHandler)
allHandlers.add('runND280CosMC','LCG', LCGRTHandler)
allHandlers.add('runND280CosMC','gLite', gLiteRTHandler)
allHandlers.add('runND280CosMC','TestSubmitter', RTHandler)
allHandlers.add('runND280CosMC','Interactive', RTHandler)
allHandlers.add('runND280CosMC','Batch', RTHandler)
allHandlers.add('runND280CosMC','Cronus', RTHandler)
allHandlers.add('runND280CosMC','Remote', LCGRTHandler)
allHandlers.add('runND280CosMC','CREAM', LCGRTHandler)




from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Adapters.IBackend import IBackend
from GangaCore.GPIDev.Adapters.IApplication import IApplication
from GangaCore.GPIDev.Adapters.IRuntimeHandler import IRuntimeHandler
from GangaCore.GPIDev.Adapters.ApplicationRuntimeHandlers import allHandlers
from GangaCore.GPIDev.Schema import *

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()

class StoreTestApplication(IApplication):
    _schema = Schema(Version(1,0), {'factor':SimpleItem(defvalue=10),'shared_config':SimpleItem(defvalue=0,protected=True,copyable=False), 'specific_config' : SimpleItem(defvalue=0,protected=True,copyable=False)} )

    _category = 'applications'
    _name = 'StoreTestApplication'

    def configure(self, masterappconfig):
        self.specific_config = self.factor*masterappconfig
        logger.debug("application.configure => "+str(self.specific_config))
        return (1,self.specific_config)

    def master_configure(self):
        self.shared_config = self.factor*10
        logger.debug("application.master_configure => "+str(self.shared_config))
        return (1,self.shared_config)
    

class StoreTestSplitter(ISplitter):
    _schema = Schema(Version(1,0), {'n':SimpleItem(defvalue=5) } )

    _category = 'splitters'
    _name = 'StoreTestSplitter'

    def split(self,job):
        from GangaCore.GPIDev.Lib.Job import Job
        subjobs = []
        for i in range(self.n):
            #print "*"*80
            j = self.createSubjob(job)
            j.application.factor = i
            subjobs.append(j)
            #import sys
            #j.printTree(sys.stdout)
        return subjobs

# This backend just stores to job config objects.

class StoreTestBackend(IBackend):
    _schema = Schema(Version(1,0), {'shared_config' : SimpleItem(defvalue=0,protected=True,copyable=True),
                                    'specific_config' : SimpleItem(defvalue=0,protected=True,copyable=True),
                                    })

    _category = 'backends'
    _name = 'StoreTestBackend'

    def submit(self,jobconfig,master_input_sandbox):
        #self.shared_config = masterjobconfig.value
        self.specific_config = jobconfig.value
        #print "backend.submit",self.shared_config,self.specific_config
        return True

    def kill(self):
        return True

    @staticmethod
    def updateMonitoringInformation(jobs):
        pass


class JobConfigObject:
    def __init__(self,value):
        self.value = value

    def __str__(self):
        return str(self.value)

    def __repr__(self):
        return self.__str__()
    
    def getSandboxFiles(self):
        return []
    
# This runtime handler just passes on the application configuration objects (adding a necessary wrapper around)
class NeutralRuntimeHandler(IRuntimeHandler):
    def master_prepare(self, app, appmasterconfig):
        r = appmasterconfig-5
        return JobConfigObject(r)

    def prepare(self,app,appsubconfig,appmasterconfig,jobmasterconfig):
        r = appsubconfig+jobmasterconfig.value-appmasterconfig
        return JobConfigObject(r)

allHandlers.add('StoreTestApplication','StoreTestBackend',NeutralRuntimeHandler)



""" Implementation of the Ganga Robot IAction class
to install ganga on a remote machine
"""

from GangaRobot.Framework.Action import IAction
from GangaRobot.Framework import Utility
from GangaCore.Utility.logging import getLogger
import GangaCore.Utility.Config
from GangaRobot.Framework.exceptions import *

import os, urllib.request, urllib.parse, urllib.error, datetime, time
from os.path import join

logger = getLogger()

class Installer(IAction):

    """ Ganga Install implementation

    has:

    > Overridden 'execute' method which does:
    >> Retrieve config
    >> Runs the ganga install script, with the config options

    """

    downloadURL = ''
    VersionNumber = ''
    VersionTime = ''
    InstallPath = ''

    def execute(self, runid):
        """
        Call the class methods in turn
        """
        #raise GangaRobotBreakError("Testing checking time",ValueError)
        Datalist = []
        
        self._getConfigInfo()

        logger.info("Running install script")
        self._Install()
        #update last download time


    def _getConfigInfo(self):
        """ Gets the config info from the ganga config file """
        config = GangaCore.Utility.Config.getConfig('TestRobot')     
        self.InstallPath = config['InstallPath']
        self.downloadURL = config['ReleasePath']        
        self.VersionNumber = config['VersionNumber']

    def _Install(self):

        """ Calls the ganga installation script """
        try: 
            f, inf = urllib.request.urlretrieve("http://ganga.web.cern.ch/ganga/download/ganga-install")
        except IOError as e:
            raise GangaRobotBreakError(e, IOError)
            
        config = GangaCore.Utility.Config.getConfig('IndependantTest')

        cmd = "python "+f+" --prefix="+self.InstallPath+" --extern=GangaTest "+self.VersionNumber
        logger.debug("Executing command: '%s'",cmd)
        try:
            os.system(cmd)
            logger.debug('Ganga installed')

        except Exception as e:
            logger.error('ganga installation failed')
            raise GangaRobotBreakError(Exception, e) # breaks the run and starts again
            



from flask import Flask
from GangaGUI.config import Config
import sys

# Disable development server warning (as it is not applicable to our GUI Flask App)
cli = sys.modules['flask.cli']
cli.show_server_banner = lambda *x: None

# Flask App and get configuration from config.py file
app = Flask(__name__)
app.config.from_object(Config)

from GangaGUI.gui import routes



#!/usr/bin/python
import os
import requests
import commands
import shutil
msg = ""

# put in a massive try..except block so I can get some feedback
try:
    # Request the list of releases
    resp = requests.get('https://api.github.com/repos/ganga-devs/ganga/releases')
    if resp.status_code == 200:
        # assume the first is the latest
        rel_ver = resp.json()[0]['name']

        # Now check if it's installed
        if rel_ver not in  os.listdir('/cvmfs/ganga.cern.ch/Ganga/install/'):
            # we need to install it
            msg = commands.getstatusoutput("/home/cvganga/ganga-cvmfs-install.sh %s" % rel_ver)[1]
            shutil.copy2('/cvmfs/ganga.cern.ch/Ganga/install/%s/lib/python3.6/site-packages/ganga/GangaRelease/tools/ganga-cvmfs-install.sh' % rel_ver, '~/')
            shutil.copy2('/cvmfs/ganga.cern.ch/Ganga/install/%s/lib/python3.6/site-packages/ganga/GangaRelease/tools/ganga-cvmfs-install-dev.sh' % rel_ver, '~/')

    else:
        msg = "Problem getting the list of releases"

except Exception as err:
    msg = str(err)

# send any messages
if msg:
    os.system('echo "%s" | mail -s "Message from CVMFS Installer" "project-ganga-developers@cern.ch"' % msg)



import GangaTest.Lib.StoreTestFramework
import GangaTest.Lib.CrashTest
import GangaTest.Lib.TestSubmitter.TestSubmitter
import GangaTest.Lib.TestSubmitter.TestSplitter
import GangaTest.Lib.TestApplication.TestApplication
import GangaTest.Lib.GListApp.GListApp
import GangaTest.Lib.TFile.TFile




##########################################################################
# Ganga Project. https://github.com/ganga-devs/ganga
#
##########################################################################
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.GPIDev.Adapters.IVirtualization import IVirtualization
from GangaCore.GPIDev.Adapters.IGangaFile import IGangaFile

class Singularity(IVirtualization):

    """
    The Singularity class can be used for either Singularity or Docker images. 
    It requires that singularity is installed on the worker node.

    For Singularity images you provide the image name and tag from Singularity 
    hub like

      j=Job()
      j.application=Executable(exe=File('my/full/path/to/executable'))
      j.virtualization = Singularity("shub://image:tag")

    Notice how the executable is given as a `File` object. This ensures that it 
    is copied to the working directory and thus will be accessible inside the 
    container.

    The container can also be provided as a Docker image from a repository. The 
    default repository is Docker hub.

      j.virtualization = Singularity("docker://gitlab-registry.cern.ch/lhcb-core/lbdocker/centos7-build:v3")

      j.virtualization = Singularity("docker://fedora:latest")   

    Another option is to provide a `GangaFile` Object which points to a 
    singularity file. In that case the singularity image file will be copied to 
    the worker node. The first example is with an image located on some shared 
    disk. This will be effective for running on a local backend or a batch 
    system with a shared disk system.

      imagefile = SharedFile('myimage.sif', locations=['/my/full/path/myimage.sif'])
      j.virtualization = Singularity(image= imagefile)

    while a second example is with an image located in the Dirac Storage 
    Element. This will be effective when using the Dirac backend.

      imagefile = DiracFile('myimage.sif', lfn=['/some/lfn/path'])
      j.virtualization = Singularity(image= imagefile)

    If the image is a private image, the username and password of the deploy 
    token can be given like the example below. Look inside Gitlab setting for 
    how to set this up. The token will only need access to the images and 
    nothing else.

      j.virtualization.tokenuser = 'gitlab+deploy-token-123'
      j.virtualization.tokenpassword = 'gftrh84dgel-245^ghHH'

    Directories can be mounted from the host to the container using key-value 
    pairs to the mounts option. If the directory is not available on the host, a 
    warning will be written to stderr of the job and no mount will be attempted.

      j.virtualization.mounts = {'/cvmfs':'/cvmfs'}

    By default the container is started in singularity with the `--nohome` 
    option. Extra options can be provided through the `options` attribute. See 
    the Singularity documentation for what is possible.

    If the singularity binary is not available in the PATH on the remote node - or has a different name, 
    it is possible to give the name of it like

      j.virtualization.binary='/cvmfs/oasis.opensciencegrid.org/mis/singularity/current/bin/singularity'

    """
    _name = 'Singularity'
    _schema = IVirtualization._schema.inherit_copy()
    _schema.datadict['image'] = SimpleItem(defvalue="",
                                           typelist=[str,'GangaCore.GPIDev.Adapters.IGangaFile.IGangaFile'],
                                           doc='Link to the container image. This can either be a singularity URL or a GangaFile object')
    _schema.datadict['binary'] = SimpleItem(defvalue="singularity",
                                            typelist=[str],
                                            doc='The virtualization binary itself. Can be an absolute path if required.')
    
    def modify_script(self, script, sandbox=False):
        """Overides parent's modify_script function
                    Arguments other than self:
                       script - Script that need to be modified

                    Return value: modified script"""
        
        if isinstance(self.image, IGangaFile):
            extra = 'virtualization_image = ' + repr(self.image.namePattern) + '\n'
        else:
            extra = 'virtualization_image = ' + repr(self.image) + '\n'
        extra = extra + 'virtualization_user = ' + repr(self.tokenuser) + '\n'
        extra = extra + 'virtualization_password = ' + repr(self.tokenpassword) + '\n'
        extra = extra + 'virtualization_mounts = ' + repr(self.mounts) + '\n'
        extra = extra + 'virtualization_options = ' + repr(self.options) + '\n'
        extra = extra + 'virtualization_binary = ' + repr(self.binary) + '\n'

        extra = extra + """
print("Using singularity")

import stat
if not ( ('XDG_RUNTIME_DIR' in runenv) and
         os.path.isdir(runenv['XDG_RUNTIME_DIR']) and
         (stat.S_IMODE(os.stat(runenv['XDG_RUNTIME_DIR']).st_mode) == 0o700) and
         os.access(runenv['XDG_RUNTIME_DIR'], os.W_OK) ):
    os.mkdir('.xdg', 0o700)
    runenv['XDG_RUNTIME_DIR'] = os.path.join(os.getcwd(), '.xdg')
           
options = []
if virtualization_user:
    runenv["SINGULARITY_DOCKER_USERNAME"] = virtualization_user
    runenv["SINGULARITY_DOCKER_PASSWORD"] = virtualization_password
for k,v in virtualization_mounts.items():
    if os.path.isdir(k):
        options = options + ['--bind' , k + ':' + v]
    else:
        print('Requested directory %s is not available and no bind will be made to container' % k)
options = options + virtualization_options
if execmd[0].startswith('./'): 
    execmd[0] = "/work_dir/"+execmd[0]
"""

        if sandbox:
            extra = extra + """
runenv['SINGULARITY_CACHEDIR']=os.path.join(os.getcwd(),'.singularity','cache')
for i in range(3):
    try:
        buildcommand = [virtualization_binary, 'build', '--sandbox', 'singularity_sandbox' , virtualization_image]
        rc = subprocess.call(buildcommand, env=runenv, shell=False)
        if rc==0:
            break
    except Exception as x:
        print('Exception occured in downloading Singularity image: ' + str(buildcommand))
        print('Err was: ' + str(x))
execmd = [virtualization_binary, '-q', 'exec', '--bind', 
          workdir+":"+"/work_dir", "--no-home"] + options + ['singularity_sandbox'] + execmd   
"""
        else:
            extra = extra + """
execmd = [virtualization_binary, '-q', 'exec', '--bind',
          workdir+":"+"/work_dir", "--no-home"] + options + [virtualization_image] + execmd   

"""
        script = script.replace('###VIRTUALIZATION###',extra)
        return script



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: ArgSplitter.py,v 1.1 2008-07-17 16:40:59 moscicki Exp $
###############################################################################

import copy
from GangaCore.Core.exceptions import SplitterError
from GangaCore.GPIDev.Adapters.ISplitter import ISplitter
from GangaCore.GPIDev.Base.Proxy import stripProxy
from GangaCore.GPIDev.Schema import Schema, Version, SimpleItem
from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList
from GangaCore.Utility.logging import getLogger
logger = getLogger()


class ArgSplitter(ISplitter):

    """
    Split job by changing the args attribute of the application.

    This splitter only applies to the applications which have args attribute (e.g. Executable, Root), or those
    with extraArgs (GaudiExec). If an application has both, args takes precedence.
    It is a special case of the GenericSplitter.

    This splitter allows the creation of a series of subjobs where
    the only difference between different jobs are their
    arguments. Below is an example that executes a ROOT script ~/analysis.C

    void analysis(const char* type, int events) {
      std::cout << type << "  " << events << std::endl;
    }

    with 3 different sets of arguments.

    s = ArgSplitter(args=[['AAA',1],['BBB',2],['CCC',3]])
    r = Root(version='5.10.00',script='~/analysis.C')
    j.Job(application=r, splitter=s)

    Notice how each job takes a list of arguments (in this case a list
    with a string and an integer). The splitter thus takes a list of
    lists, in this case with 3 elements so there will be 3 subjobs.

    Running the subjobs will produce the output:
    subjob 1 : AAA  1
    subjob 2 : BBB  2
    subjob 3 : CCC  3
"""
    _name = "ArgSplitter"
    _schema = Schema(Version(1, 0), {
        'args': SimpleItem(defvalue=[], typelist=[list, GangaList], sequence=1, doc='A list of lists of arguments to pass to script')
    })

    def split(self, job):

        subjobs = []

        for arg in self.args:
            j = self.createSubjob(job,['application'])
            # Add new arguments to subjob
            app = copy.deepcopy(job.application)
            if hasattr(app, 'args'):
                app.args = arg
            elif hasattr(app, 'extraArgs'):
                app.extraArgs = arg
            else:
                raise SplitterError('Application has neither args or extraArgs in its schema') 
                    
            j.application = app
            logger.debug('Arguments for split job is: ' + str(arg))
            subjobs.append(stripProxy(j))

        return subjobs


#
#
# $Log: not supported by cvs2svn $
# Revision 1.7.4.3  2008/07/03 08:36:16  wreece
# Typesystem fix for Splitters
#
# Revision 1.7.4.2  2008/03/12 12:42:38  wreece
# Updates the splitters to check for File objects in the list
#
# Revision 1.7.4.1  2008/02/08 15:09:52  amuraru
# fixed the TypeMismatchError
#
# Revision 1.7  2007/07/10 13:08:32  moscicki
# docstring updates (ganga devdays)
#
# Revision 1.6  2006/09/29 08:31:56  moscicki
# typo fix
#
# Revision 1.5  2006/09/15 14:24:24  moscicki
# fixed a typo
#
# Revision 1.4  2006/08/24 16:52:10  moscicki
# splitter.createJob()
#
# Revision 1.3  2006/08/03 08:14:30  moscicki
# SimpleItem fix
#
# Revision 1.2  2006/08/01 10:25:50  moscicki
# small schema fixes
#
# Revision 1.1  2006/06/21 11:27:29  moscicki
# splitters moved to the new location
#
#
#



from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.logging import getLogger

from GangaCore.Utility.GridShell import getShell

logger = getLogger()

logger.critical('LCG Grid Simulator ENABLED')

##########################################################################
# GRID SIMULATOR
##########################################################################

config = getConfig("GridSimulator")

def sleep(val):
    import time
    time.sleep(get_number(val))


def failed(val):
    t = get_number(val)
    import random
    return random.random() < t


def get_number(val):
    import random
    if isinstance(val, str):
        t = eval(val, {'random': random})
    else:
        t = val
    if not type(t) in [type(1.0), type(1)]:
        # print 'problem with configuration option, invalid value: %s'%val
        logger.error(
            'problem with configuration option, invalid value: %s', val)
        return 0
    # print t
    return t

import os
import time

cmd = 'simulation'


class GridSimulator(object):

    '''Simulator of LCG interactions'''

    credential = None

    def __init__(self):
        self.active = True
        #import GangaCore.Core.FileWorkspace
        #basedir = GangaCore.Core.FileWorkspace.gettop()
        #basedir = '/tmp'
        basedir = '.'
        self.gridmap_filename = '%s/lcg_simulator_gridmap' % basedir
        import shelve
        # map Grid job id into inputdir (where JDL file is)
        self.jobid_map = shelve.open(self.gridmap_filename, writeback=False)
        self.jobid_map.setdefault('_job_count', 0)

        # here we store the job finish times as seen by ganga
        self.finished_jobs_filename = '%s/lcg_simulator_finished_jobs' % basedir
        self.ganga_finish_time = shelve.open(
            self.finished_jobs_filename, writeback=False)

        self.shell = getShell('GLITE')

        logger.critical('Grid Simulator data files: %s %s',
                        self.gridmap_filename, self.finished_jobs_filename)

    def check_proxy(self):
        return True

    def submit(self, jdlpath, ce=None):
        '''This method is used for normal and native bulk submission supported by GLITE middleware.'''

        logger.debug(
            'job submit command: submit(jdlpath=%s,ce=%s)', jdlpath, ce)

        jdl = eval(file(jdlpath).read())

        subjob_ids = []
        if jdl['Type'] == 'collection':
            import re
            # we need to parse the Nodes attribute string here
            r = re.compile(r'.*NodeName = "(gsj_\d+)"; file="(\S*)"*')
            for line in jdl['Nodes'].splitlines()[1:-1]:
                m = r.match(line)
                if m:
                    nodename, sjdl_path = m.groups()
                subjob_ids.append(
                    self._submit(sjdl_path, ce, [], nodename=nodename))

        masterid = self._submit(jdlpath, ce, subjob_ids)

        return masterid

    def _params_filename(self, jobid):
        inputdir = os.path.realpath(self.jobid_map[jobid])
        return os.path.join(inputdir, 'params')

    def _submit(self, jdlpath, ce, subjob_ids, nodename=None):
        '''Submit a JDL file to LCG'''

        logger.debug(
            'job submit command: _submit(jdlpath=%s,ce=%s,subjob_ids=%s)', jdlpath, ce, subjob_ids)

        inputdir = os.path.dirname(os.path.realpath(jdlpath))

        def write():
            with open(os.path.join(inputdir, 'params'), 'w') as file_:
                file_.write(repr(runtime_params))

        runtime_params = {}
        runtime_params['submission_time_start'] = time.time()

        sleep(config['submit_time'])
        runtime_params['submission_time_stop'] = time.time()

        if failed(config['submit_failure_rate']):
            runtime_params['status'] = 'failed_to_submit'
            write()
            logger.warning('Job submission failed.')
            return

        jobid = self._make_new_id()

        self.jobid_map[jobid] = inputdir

        runtime_params['jobid'] = jobid
        runtime_params['status'] = 'submitted'
        runtime_params['should_fail'] = failed(config['job_failure_rate'])
        runtime_params['expected_job_id_resolve_time'] = get_number(
            config['job_id_resolved_time'])
        runtime_params['expected_finish_time'] = time.time(
        ) + get_number(config['job_finish_time'])
        runtime_params['subjob_ids'] = subjob_ids
        runtime_params['nodename'] = nodename
        write()
        return jobid

    def _make_new_id(self):
        self.jobid_map['_job_count'] += 1
        jobid = 'https://ganga.simulator.cern.ch/%d' % self.jobid_map[
            '_job_count']
        return jobid

    def _cancel(self, jobid):
        inputdir = self.jobid_map[jobid]

        sleep(config['cancel_time'])
        if failed(config['cancel_failure_rate']):
            with open(self._params_filename(jobid), 'a') as file_:
                file_.write('\n failed to cancel: %d' % time.time())
            return False
        with open(self._params_filename(jobid), 'a') as file_:
            file_.write('\ncancelled: %d' % time.time())
        return True

    def native_master_cancel(self, jobid):
        '''Native bulk cancellation supported by GLITE middleware.'''

        logger.debug(
            'job cancel command: native_master_cancel(jobid=%s', jobid)

        # FIXME: TODO: emulate bulk!
        return self._cancel(jobid)

    def _status(self, jobid, has_id):
        logger.debug(
            'job status command: _status(jobid=%s,has_id=%d)', jobid, has_id)

        info = {'id': None,
                'name': None,
                'status': None,
                'exit': None,
                'reason': None,
                'is_node': False,
                'destination': 'anywhere'}

        params = eval(file(self._params_filename(jobid)).read())

        sleep(config['single_status_time'])

        assert params['jobid'] == jobid

        if has_id:
            info['id'] = params['jobid']
            info['name'] = params['nodename']

        # if is_collection and time.time() > params['expected_job_id_resolve_time']:
        #    info['name'] = 'node_%d' % 0 # FIXME: some number (read from jdl?)

        logger.debug('current_time-expected_finish_time = %d',
                     time.time() - params['expected_finish_time'])

        if time.time() > params['expected_finish_time']:
            if params['should_fail']:
                info['status'] = 'Aborted'
                info['reason'] = 'for no reason'
                info['exit'] = -1
                self.ganga_finish_time[jobid] = time.time()
            else:
                info['status'] = 'Done (Success)'
                info['exit'] = 0
                info['reason'] = 'for a reason'

        logger.debug('_status (jobid=%s) -> %s', jobid, repr(info))

        # PENDING: handle other statuses: 'Running','Aborted','Cancelled','Done
        # (Exit Code !=0)','Cleared'
        return info

    def status(self, jobids, is_collection=False):
        '''Query the status of jobs on the grid.
        If is_collection is False then jobids is a list of non-split jobs or emulated bulk subjobs of a single master job.
        If is_collection is True then jobids is a list of master jobs which are natively bulk.
        '''

        logger.debug(
            'job status command: status(jobid=%s,is_collection=%d)', jobids, is_collection)

        info = []

        for id in jobids:
            if is_collection:
                # print 'master _status'
                sleep(config['master_status_time'])
                info.append(self._status(id, True))
                # print 'master _status done'
                params = eval(file(self._params_filename(id)).read())
                # print 'master params',params
                has_id = time.time() > params['expected_job_id_resolve_time']
                for sid in params['subjob_ids']:
                    info.append(self._status(sid, has_id))
                    info[-1]['is_node'] = True
            else:
                has_id = False
                info.append(self._status(id, True))

        return info

    def get_loginfo(self, jobid, directory, verbosity=1):
        '''Fetch the logging info of the given job and save the output in the jobs outputdir'''

        return ""

    def get_output(self, jobid, directory, wms_proxy=False):
        '''Retrieve the output of a job on the grid'''

        logger.debug(
            'job get output command: get_output(jobid=%s,directory=%s)', jobid, directory)
        sleep(config['get_output_time'])
        self.ganga_finish_time[jobid] = time.time()
        return (True, None)

    def cancel(self, jobid):
        '''Cancel a job'''
        logger.debug('job cancel command: cancel(jobid=%s)', jobid)

        return self._cancel(jobid)

    @staticmethod
    def expandjdl(items):
        '''Expand jdl items'''

        return repr(items)



import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger()

logger.debug("Loading IApplication")
import GangaCore.GPIDev.Adapters.IApplication
logger.debug("Loading IBackend")
import GangaCore.GPIDev.Adapters.IBackend
logger.debug("Loading ISplitter")
import GangaCore.GPIDev.Adapters.ISplitter
logger.debug("Loading IMerger")
import GangaCore.GPIDev.Adapters.IMerger

logger.debug("Loading GangaList")
import GangaCore.GPIDev.Lib.GangaList
logger.debug("Loading File")
import GangaCore.GPIDev.Lib.File
logger.debug("Loading Job")
import GangaCore.GPIDev.Lib.Job

logger.debug("Loading Mergers")
import GangaCore.Lib.Mergers
logger.debug("Loading Splitters")
import GangaCore.Lib.Splitters

logger.debug("Loading Executable")
import GangaCore.Lib.Executable
logger.debug("Loading Root")
import GangaCore.Lib.Root
logger.debug("Loading Notebook")
import GangaCore.Lib.Notebook

logger.debug("Loading LocalHost")
import GangaCore.Lib.Localhost
logger.debug("Loading LCG")
import GangaCore.Lib.LCG
logger.debug("Loading Condor")
import GangaCore.Lib.Condor
logger.debug("Loading Interactive")
import GangaCore.Lib.Interactive
logger.debug("Loading Batch")
import GangaCore.Lib.Batch
logger.debug("Loading Remote")
import GangaCore.Lib.Remote

logger.debug("Loading Tasks")
import GangaCore.GPIDev.Lib.Tasks


logger.debug("Loading Checkers")
import GangaCore.Lib.Checkers
logger.debug("Loading Notifier")
import GangaCore.Lib.Notifier

logger.debug("Loading virtualization classes")
import GangaCore.Lib.Virtualization

logger.debug("Finished Runtime.plugins")



# Dirac commands
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/

@diracCommand
def getJobGroupJobs(jg):
    ''' Return jobs in a group'''
    return dirac.selectJobs(jobGroup=jg)


@diracCommand
def kill(id):
    ''' Kill a given DIRAC Job ID within DIRAC '''
    return dirac.deleteJob(id)


@diracCommand
def peek(id):
    ''' Peek at the DIRAC Job id and return what we saw '''
    return dirac.peekJob(id)


@diracCommand
def getJobCPUTime(id):
    ''' Get the amount of CPU time taken by the DIRAC Job id'''
    return dirac.getJobCPUTime(id)


@diracCommand
def reschedule(id):
    ''' Reschedule within DIRAC a given DIRAC Job id'''
    return dirac.reschedule(id)


@diracCommand
def submit(djob, mode='wms'):
    ''' Submit a DIRAC job given by the jdl:djob with a given mode '''
    return dirac.submitJob(djob, mode=mode)


@diracCommand
def ping(system, service):
    ''' Ping a given service on a given system running DIRAC '''
    return dirac.ping(system, service)


@diracCommand
def removeFile(lfn):
    ''' Remove a given LFN from the DFC'''
    ret = {}
    if type(lfn) is list:
        for l in lfn:
            ret.update(dirac.removeFile(l))
    else:
        ret.update(dirac.removeFile(lfn))
    return ret


@diracCommand
def getMetadata(lfn):
    ''' Return the metadata associated with a given :DN'''
    return dirac.getLfnMetadata(lfn)


@diracCommand
def getReplicas(lfns):
    ''' Return  the locations of the replicas of a given LFN in a dict format, SE: location '''
    return dirac.getReplicas(lfns, active=True, preferDisk = True)

@diracCommand
def getReplicasForJobs(lfns):
    ''' Return the locations of the replicas of a given LFN in a dict format, SE: location.
        This is for use in the splitter to negate copies at SEs that are not to be used for user jobs '''
    return dirac.getReplicasForJobs(lfns)

@diracCommand
def getAccessURL(lfn, SE, protocol=False):
    ''' Return the access URL for the given LFN, storage element and protocol. The protocol should be in the form of a list '''
    return dirac.getAccessURL(lfn, SE, False, protocol)


@diracCommand
def getFile(lfns, destDir=''):
    ''' Put the physical file behind the LFN in the destDir path'''
    return dirac.getFile(lfns, destDir=destDir)


@diracCommand
def replicateFile(lfn, destSE, srcSE='', locCache=''):
    ''' Replicate a given LFN from a srcSE to a destSE'''
    res = dirac.replicateFile(lfn, destSE, srcSE, locCache)
    return res


@diracCommand
def removeReplica(lfn, sE):
    ''' Remove the physical files and LFN from the DFC'''
    return dirac.removeReplica(lfn, sE)


@diracCommand
def getOutputData(id, outputFiles='', destinationDir=''):
    ''' Return output data of a requeted DIRAC Job id, place outputFiles in a given destinationDir') '''
    return dirac.getJobOutputData(id, outputFiles, destinationDir)


@diracCommand
def splitInputData(files, files_per_job):
    ''' Split list of files ito a list of list of smaller files (below files_per_job in length) and return the list of lists'''
    return dirac.splitInputData(files, files_per_job)


@diracCommand
def getInputDataCatalog(lfns, site, xml_file):
    ''' Get the XML describing the given LFNs at a given site'''
    return dirac.getInputDataCatalog(lfns, site, xml_file)


@diracCommand
def uploadFile(lfn, file, diracSEs, guid=None):
    ''' Upload a given file to an lfn with 1 replica places at each element in diracSEs. Use a given guid if given'''
    outerr = {}
    for se in diracSEs:
        result = dirac.addFile(lfn, file, se, guid)
        if result.get('OK', False) and lfn in result.get('Value', {'Successful': {}})['Successful']:
            result['Value']['Successful'][lfn].update({'DiracSE': se})
            md = dirac.getLfnMetadata(lfn)
            if md.get('OK', False) and lfn in md.get('Value', {'Successful': {}})['Successful']:
                guid = md['Value']['Successful'][lfn]['GUID']
                result['Value']['Successful'][lfn].update({'GUID': guid})
            return result
        outerr.update({se: result})
    
    return outerr


@diracCommand
def addFile(lfn, file, diracSE, guid):
    ''' Upload a given file to an lfn with 1 replica places at each element in diracSEs. Use a given guid if given'''
    return dirac.addFile(lfn, file, diracSE, guid)


@diracCommand
def getOutputSandbox(id, outputDir=os.getcwd(), unpack=True, oversized=True, noJobDir=True, pipe_out=True):
    '''
    Get the outputsandbox and return the output from Dirac to the calling function
    id: the DIRAC jobid of interest
    outputDir: output directory locall on disk to use
    oversized: is this output sandbox oversized this will be modified
    noJobDir: should we create a folder with the DIRAC job ID?
    output: should I output the Dirac output or should I return a python object (False)
    unpack: should the sandbox be untarred when downloaded'''
    result = dirac.getOutputSandbox(id, outputDir, oversized, noJobDir, unpack)
    if result is not None and result.get('OK', False):

        if not noJobDir:
            tmpdir = os.path.join(outputDir, str(id))
            os.system('mv -f %s/* %s/. ; rm -rf %s' % (tmpdir, outputDir, tmpdir))
        
        os.system('for file in $(ls %s/*Ganga_*.log); do ln -s ${file} %s/stdout; break; done' % (outputDir, outputDir))
    #So the download failed. Maybe the sandbox was oversized and stored on the grid. Check in the job parameters and download it
    else:
        parameters = dirac.getJobParameters(id)
        if parameters is not None and parameters.get('OK', False):
            parameters = parameters['Value']
            if 'OutputSandboxLFN' in parameters:
                result = dirac.getFile(parameters['OutputSandboxLFN'], destDir=outputDir)
                dirac.removeFile(parameters['OutputSandboxLFN'])
    return result


@diracCommand
def getOutputDataInfo(id, pipe_out=True):
    ''' Get information on the output data generated by a job of ID and pipe it out or return it'''
    ret = {}
    result = getOutputDataLFNs(id, pipe_out=False)
    if result.get('OK', False) and 'Value' in result:
        for lfn in result.get('Value', []):
            file_name = os.path.basename(lfn)
            ret[file_name] = {}
            ret[file_name]['LFN'] = lfn
            md = dirac.getLfnMetadata(lfn)
            if md.get('OK', False) and lfn in md.get('Value', {'Successful': {}})['Successful']:
                ret[file_name]['GUID'] =  md['Value']['Successful'][lfn]['GUID']
            # this catches if fail upload, note lfn still exists in list as
            # dirac tried it
            elif md.get('OK', False) and lfn in md.get('Value', {'Failed': {}})['Failed']:
                ret[file_name]['LFN'] = '###FAILED###'
                ret[file_name]['LOCATIONS'] = md['Value']['Failed'][lfn]
                ret[file_name]['GUID'] = 'NotAvailable'
                continue
            rp = dirac.getReplicas(lfn)
            if rp.get('OK', False) and lfn in rp.get('Value', {'Successful': {}})['Successful']:
                ret[file_name]['LOCATIONS'] = rp['Value']['Successful'][lfn].keys()
    return ret



# could shrink this with dirac.getJobOutputLFNs from ##dirac
@diracCommand
def getOutputDataLFNs(id, pipe_out=True):
    ''' Get the outputDataLFN which have been generated by a Dirac job of ID and pipe it out or return it'''
    parameters = dirac.getJobParameters(id)
    lfns = []
    ok = False
    message = 'The outputdata LFNs could not be found.'

    if parameters is not None and parameters.get('OK', False):
        parameters = parameters['Value']
        # remove the sandbox if it has been uploaded
        sandbox = None
        if 'OutputSandboxLFN' in parameters:
            sandbox = parameters['OutputSandboxLFN']

        # now find out about the outputdata
        if 'UploadedOutputData' in parameters:
            lfn_list = parameters['UploadedOutputData']
            import re
            lfns = re.split(',\s*', lfn_list)
            if sandbox is not None and sandbox in lfns:
                lfns.remove(sandbox)
            ok = True
        elif parameters is not None and 'Message' in parameters:
            message = parameters['Message']

    result = {'OK': ok}
    if ok:
        result['Value'] = lfns
    else:
        result['Message'] = message

    return result


@diracCommand
def normCPUTime(id, pipe_out=True):
    ''' Get the normalied CPU time that has been used by a DIRAC job of ID and pipe it out or return it'''
    parameters = dirac.getJobParameters(id)
    ncput = None
    if parameters is not None and parameters.get('OK', False):
        parameters = parameters['Value']
        if 'NormCPUTime(s)' in parameters:
            ncput = parameters['NormCPUTime(s)']
    return ncput


@diracCommand
def finished_job(id, outputDir=os.getcwd(), unpack=True, oversized=True, noJobDir=True, downloadSandbox = True):
    ''' Nesting function to reduce number of calls made against DIRAC when finalising a job, takes arguments such as getOutputSandbox
    Returns the CPU time of the job as a dict, the output sandbox information in another dict and a dict of the LFN of any uploaded data'''
    out_cpuTime = normCPUTime(id, pipe_out=False)
    if downloadSandbox:
        out_sandbox = getOutputSandbox(id, outputDir, unpack, oversized, noJobDir, pipe_out=False)
    else:
        out_sandbox = None
    out_dataInfo = getOutputDataInfo(id, pipe_out=False)
    outStateTime = {'completed' : getStateTime(id, 'completed', pipe_out=False)}
    return (out_cpuTime, out_sandbox, out_dataInfo, outStateTime)


@diracCommand
def finaliseJobs(inputDict, statusmapping, downloadSandbox=True, oversized=True, noJobDir=True):
    ''' A function to get the necessaries to finalise a whole bunch of jobs. Returns a dict of job information and a dict of stati.'''
    returnDict = {}
    statusList = dirac.getJobStatus(inputDict.keys())
    for diracID in inputDict.keys():
        returnDict[diracID] = {}
        returnDict[diracID]['cpuTime'] = normCPUTime(diracID, pipe_out=False)
        if downloadSandbox:
            returnDict[diracID]['outSandbox'] = getOutputSandbox(diracID, inputDict[diracID], oversized, noJobDir, pipe_out=False)
        else:
            returnDict[diracID]['outSandbox'] = None
        returnDict[diracID]['outDataInfo'] = getOutputDataInfo(diracID, pipe_out=False)
        returnDict[diracID]['outStateTime'] = {'completed' : getStateTime(diracID, 'completed', pipe_out=False)}
    return returnDict, statusList


@diracCommand
def status(job_ids, statusmapping, pipe_out=True):
    '''Function to check the statuses and return the Ganga status of a job after looking it's DIRAC status against a Ganga one'''
    # Translate between the many statuses in DIRAC and the few in Ganga

    #return {'OK':True, 'Value':[['WIP', 'WIP', 'WIP', 'WIP', 'WIP']]}

    result = dirac.getJobStatus(job_ids)
    if not result['OK']:
        return result
    status_list = []
    bulk_status = result['Value']
    for _id in job_ids:
        job_status = bulk_status.get(_id, {})
        minor_status = job_status.get('MinorStatus', None)
        dirac_status = job_status.get('Status', None)
        dirac_site = job_status.get('Site', None)
        ganga_status = statusmapping.get(dirac_status, None)
        if ganga_status is None:
            ganga_status = 'failed'
            dirac_status = 'Unknown: No status for Job'
        #if dirac_status == 'Completed' and (minor_status not in ['Pending Requests']):
        #    ganga_status = 'running'
        if minor_status in ['Uploading Output Data']:
            ganga_status = 'running'

        try:
            from DIRAC.Core.DISET.RPCClient import RPCClient
            monitoring = RPCClient('WorkloadManagement/JobMonitoring')
            app_status = monitoring.getJobAttributes(_id)['Value']['ApplicationStatus']
        except:
            app_status = "unknown ApplicationStatus"

        status_list.append([minor_status, dirac_status, dirac_site, ganga_status, app_status])

    return status_list


@diracCommand
def getStateTime(id, status, pipe_out=True):
    ''' Return the state time from DIRAC corresponding to DIRACJob tranasitions'''
    log = dirac.getJobLoggingInfo(id)
    if 'Value' not in log:
        return None
    L = log['Value']
    checkstr = ''

    if status == 'running':
        checkstr = 'Running'
    elif status == 'completed':
        checkstr = 'Done'
    elif status == 'completing':
        checkstr = 'Completed'
    elif status == 'failed':
        checkstr = 'Failed'
    else:
        checkstr = ''

    if checkstr == '':
        print("%s" % None)
        return

    for l in L:
        if checkstr in l[0]:
            T = datetime.datetime(*(time.strptime(l[3], "%Y-%m-%d %H:%M:%S")[0:6]))
            return T
    
    return None


@diracCommand
def getBulkStateTime(job_ids, status, pipe_out=True):
    ''' Function to repeatedly call getStateTime for multiple Dirac Job id and return the result in a dictionary '''
    result = {}
    for this_id in job_ids:
        result[this_id] = getStateTime(this_id, status, pipe_out=False)

    return result


@diracCommand
def monitorJobs(job_ids, status_mapping, pipe_out=True):
    ''' This combines 'status' and 'getBulkStateTime' into 1 function call for monitoring
    '''
    status_info = status(job_ids, status_mapping, pipe_out=False)
    state_job_status = {}
    for job_id, this_stat_info in zip(job_ids, status_info):
        if this_stat_info:
            update_status = this_stat_info[3]
            if update_status not in state_job_status:
                state_job_status[update_status] = []
            state_job_status[update_status].append(job_id)
    state_info = {}
    for this_status, these_jobs in state_job_status.iteritems():
        state_info[this_status] = getBulkStateTime(these_jobs, this_status, pipe_out=False)

    return (status_info, state_info)


@diracCommand
def timedetails(id):
    ''' Function to return the getJobLoggingInfo for a DIRAC Job of id'''
    log = dirac.getJobLoggingInfo(id)
    d = {}
    for i in range(0, len(log['Value'])):
        d[i] = log['Value'][i]
    return d

# DiracAdmin commands
#/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/\/


@diracCommand
def getJobPilotOutput(id, dir):
    ''' Get the output of the DIRAC pilot that this job was running on and place it in dir'''
    pwd = os.getcwd()
    try:
        os.chdir(dir)
        os.system('rm -f pilot_%d/std.out && rmdir pilot_%d ' % (id, id))
        result = DiracAdmin().getJobPilotOutput(id)
    finally:
        os.chdir(pwd)
    return result


@diracCommand
def getServicePorts():
    ''' Get the service ports from the DiracAdmin based upon the Dirac config'''
    return DiracAdmin().getServicePorts()



@diracCommand
def getSitesForSE(se):
    ''' Get the Sites associated with this SE'''
    from DIRAC.Core.Utilities.SiteSEMapping import getSitesForSE
    result = getSitesForSE(storageElement=se)
    return result


@diracCommand
def getSEsForSite(site):
    ''' Get the list of SE associated with this site'''
    from DIRAC.Core.Utilities.SiteSEMapping import getSEsForSite
    result = getSEsForSite(site)
    return result


@diracCommand
def getSESiteMapping():
    '''Get the mapping of SEs and sites'''
    from DIRAC.Core.Utilities.SiteSEMapping import getSESiteMapping
    result = getSESiteMapping()
    return result


@diracCommand
def checkSEStatus(se, access = 'Write'):
    ''' returns the value of a certain SE status flag (access or other)
      param se: Storage Element name
      type se: string
      param access: type of access
      type access: string in ('Read', 'Write', 'Remove', 'Check')
       returns: True or False
    '''
    result = dirac.checkSEAccess(se, access)
    return result

@diracCommand
def listFiles(baseDir, minAge = None):
    ''' Return a list of LFNs for files stored on the grid in the argument
        directory and its subdirectories
        param baseDir: Top directory to begin search
        type baseDir: string
        param minAge: minimum age of files to be returned
        type minAge: string format: "W:D:H"
    '''

    from DIRAC.Resources.Catalog.FileCatalog import FileCatalog
    fc = FileCatalog()

    from datetime import datetime, timedelta

    withMetaData = False
    cutoffTime = datetime.utcnow()
    import re
    r = re.compile('\d:\d:\d')
    if r.match(minAge):
        withMetaData = True
        timeList = minAge.split(':')
        timeLimit = timedelta(weeks = int(timeList[0]), days = int(timeList[1]), hours = int(timeList[2]))
        cutoffTime = datetime.utcnow() - timeLimit

    baseDir = baseDir.rstrip('/')

    activeDirs = [baseDir]

    allFiles = []
    emptyDirs = []

    while len(activeDirs) > 0:
        currentDir = activeDirs.pop()	
        res = fc.listDirectory(currentDir, withMetaData, timeout = 360)
        if not res['OK']:
            return "Error retrieving directory contents", "%s %s" % ( currentDir, res['Message'] )
        elif currentDir in res['Value']['Failed']:
            return "Error retrieving directory contents", "%s %s" % ( currentDir, res['Value']['Failed'][currentDir] )
        else:
            dirContents = res['Value']['Successful'][currentDir]
            subdirs = dirContents['SubDirs']
            files = dirContents['Files']
            if not subdirs and not files:
                emptyDirs.append( currentDir )
            else:
                for subdir in sorted( subdirs, reverse=True):
                    if (not withMetaData) or subdirs[subdir]['CreationDate'] < cutoffTime:
                        activeDirs.append(subdir)
                for filename in sorted(files):
                    fileOK = False
                    if (not withMetaData) or files[filename]['MetaData']['CreationDate'] < cutoffTime:
                        fileOK = True
		    if not fileOK:
                        files.pop(filename)
                allFiles += sorted(files)

    return allFiles




#! /usr/bin/env python

"""
This script generates ReST files based on the contents of the GPI.
These files are then parsed by Sphinx to create the GPI documentation.
"""

from __future__ import print_function, absolute_import

import sys
import os
import inspect
from itertools import zip_longest
import shutil

doc_dir = os.path.dirname(os.path.realpath(__file__))
python_dir = doc_dir + '/../ganga'
sys.path.insert(0, os.path.abspath(os.path.join(doc_dir, '..', 'ganga')))

from GangaCore.GPIDev.Base.Proxy import GPIProxyObject, stripProxy, getName

print('Generating GPI documentation')

## LOADING GANGA ##
import GangaCore.PACKAGE
GangaCore.PACKAGE.standardSetup()

import GangaCore.Runtime
gangadir = os.path.expandvars('$HOME/gangadir_sphinx_dummy')
this_argv = [
    'ganga',  # `argv[0]` is usually the name of the program so fake that here
    '-o[Configuration]RUNTIME_PATH=GangaCore',
    '-o[Configuration]gangadir={gangadir}'.format(gangadir=gangadir),
]

# Actually parse the options
GangaCore.Runtime._prog = GangaCore.Runtime.GangaProgram(argv=this_argv)
GangaCore.Runtime._prog.parseOptions()

# Perform the configuration and bootstrap steps in ganga
GangaCore.Runtime._prog.configure()
GangaCore.Runtime._prog.initEnvironment()
GangaCore.Runtime._prog.bootstrap(interactive=False)
## FINISHED LOADING GANGA ##


def indent(s, depth=''):
    # type: (str, str) -> str
    """
    Adds `indent` to the beginning of every line
    """
    return '\n'.join(depth+l for l in s.splitlines())


def reindent(docstring, depth=''):
    # type: (str, str) -> str
    """
    Returns ``docstring`` trimmed and then each line prepended with ``depth``
    """
    return indent(inspect.cleandoc(docstring), depth)


def signature(func, name=None):
    """
    Args:
        func: a function object
        name: an optional name for the function in case the function has been aliased

    Returns: a string representing its signature as would be written in code

    """
    args, varargs, varkw, defaults = inspect.getargspec(func)
    defaults = defaults or []  # If there are no defaults, set it to an empty list
    defaults = [repr(d) for d in defaults]  # Type to get a useful string representing the default

    # Based on a signature like foo(a, b, c=None, d=4, *args, **kwargs)
    # we get args=['a', 'b', 'c', 'd'] and defaults=['None', '4']
    # We must match them backwards from the end and pad the beginning with None
    # to get arg_pairs=[('a', None), ('b', None), ('c', 'None'), ('d', '4')]
    arg_pairs = reversed([(a, d) for a, d in zip_longest(reversed(args), reversed(defaults), fillvalue=None)])
    # Based on arg_pairs we convert it into
    # arg_strings=['a', 'b', 'a=None', 'd=4']
    arg_strings = []
    for arg, default in arg_pairs:
        full_arg = arg
        if default is not None:
            full_arg += '='+default
        arg_strings.append(full_arg)
    # and append args and kwargs if necessary to get
    # arg_strings=['a', 'b', 'a=None', 'd=4', '*args', '**kwargs']
    if varargs is not None:
        arg_strings.append('*'+varargs)
    if varkw is not None:
        arg_strings.append('**'+varkw)

    # Signature is then 'foo(a, b, c=None, d=4, *args, **kwargs)'
    return '{name}({args})'.format(name=name or func.__name__, args=', '.join(arg_strings))


# First we get all objects that are in Ganga.GPI and filter out any non-GangaObjects
gpi_classes = [stripProxy(o) for name, o in GangaCore.GPI.__dict__.items() if isinstance(o, type) and issubclass(o, GPIProxyObject)]

with open(doc_dir+'/GPI/classes.rst', 'w') as cf:

    print('GPI classes', file=cf)
    print('===========', file=cf)
    print('', file=cf)

    # For each class we generate a set of ReST using '.. class::' etc.
    for c in gpi_classes:
        print('.. class:: {module_name}'.format(module_name=getName(c)), file=cf)

        if c.__doc__:
            print('', file=cf)
            print(reindent(c.__doc__, '    '), file=cf)

        print('', file=cf)

        print(reindent('Plugin category: ' + c._category, '    '), file=cf)

        print('', file=cf)

        # Go through each schema item and document it if it's not hidden
        visible_items = ((name, item) for name, item in c._schema.allItems() if not item['hidden'])
        for name, item in visible_items:
            # These are the only ones we want to show in the docs
            properties_we_care_about = ['protected', 'defvalue', 'changable_at_resubmit']
            props = dict((k, v) for k, v in item._meta.items() if k in properties_we_care_about)

            print('    .. attribute:: {name}'.format(name=name), file=cf)
            print('', file=cf)
            print(reindent(item['doc'], depth='        '), file=cf)
            print(reindent(str(props), depth='        '), file=cf)
            print('', file=cf)

        # Add documentation for each exported method
        for method_name in c._exportmethods:
            try:
                f = getattr(c, method_name).__func__
            except AttributeError as e:
                # Some classes have erroneous export lists so we catch them here and print a warning
                print('WARNING on class', getName(c), ':', end=' ', file=sys.stderr)
                print(e, file=sys.stderr)
                continue

            print('    .. method:: {signature}'.format(signature=signature(f)), file=cf)
            if f.__doc__:
                print('', file=cf)
                print(reindent(f.__doc__, '        '), file=cf)
                print('', file=cf)

        print('', file=cf)
        print('', file=cf)

# Looking through the plugin list helps categorise the GPI objects

with open(doc_dir+'/GPI/plugins.rst', 'w') as pf:
    from GangaCore.Utility.Plugin.GangaPlugin import allPlugins

    print('Plugins', file=pf)
    print('=======', file=pf)
    print('', file=pf)

    for c, ps in allPlugins.allCategories().items():
        print(c, file=pf)
        print('-'*len(c), file=pf)
        print('', file=pf)

        for name, c in ps.items():
            if c._declared_property('hidden'):
                continue
            print('* :class:`{name}`'.format(name=name), file=pf)
        print('', file=pf)

print('')

# All objects that are not proxied GangaObjects
gpi_objects = dict((name, stripProxy(o)) for name, o in GangaCore.GPI.__dict__.items() if stripProxy(o) not in gpi_classes and not name.startswith('__'))

# Any objects which are not exposed as proxies (mostly exceptions)
non_proxy_classes = dict((k, v) for k, v in gpi_objects.items() if inspect.isclass(v))

# Anything which is callable
callables = dict((k, v) for k, v in gpi_objects.items() if callable(v) and v not in non_proxy_classes.values())

# Things which were declared as actual functions
functions = dict((k, v) for k, v in callables.items() if inspect.isfunction(v) or inspect.ismethod(v))

with open(doc_dir+'/GPI/functions.rst', 'w') as ff:

    print('Functions', file=ff)
    print('=========', file=ff)
    print('', file=ff)

    for name, func in functions.items():

        print('.. function:: {signature}'.format(signature=signature(func, name)), file=ff)
        print('', file=ff)
        print(reindent(func.__doc__ or '', '    '), file=ff)

        print('', file=ff)

## EXITING GANGA ##
from GangaCore.Core.InternalServices import ShutdownManager

# make sure we don't have an interactive shutdown policy
from GangaCore.Core.GangaThread import GangaThreadPool
GangaThreadPool.shutdown_policy = 'batch'

# This should now be safe
ShutdownManager._protected_ganga_exitfuncs()

shutil.rmtree(gangadir, ignore_errors=True)
## FINISHED EXITING GANGA ##



###############################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: Interactive.py,v 1.1 2008-07-17 16:40:57 moscicki Exp $
###############################################################################
# File: Interactive.py
# Author: K. Harrison
# Created: 060720
# Version 1.0: 060728
#
# KH 060803 - Corrected _getJobObject to getJobObject
#
# KH 060829 - Updated to use Sandbox module
#
# KH 060901 - Updates in submit and preparejob methods, for core changes
#
# KH 061103 - Corrected to take into account master input sandbox
#
# KH 080306 - Corrections from VR

"""Module containing class for running jobs interactively"""

__author__ = "K.Harrison <Harrison@hep.phy.cam.ac.uk>"
__date__ = "6 February 2008"
__version__ = "1.4"

from GangaCore.Core import Sandbox
from GangaCore.GPIDev.Adapters.IBackend import IBackend
from GangaCore.GPIDev.Base.Proxy import stripProxy, getName
from GangaCore.GPIDev.Lib.File import FileBuffer
from GangaCore.GPIDev.Schema import Schema, SimpleItem, Version
from GangaCore.Utility import util
from GangaCore.Utility.Config import getConfig
from GangaCore.Utility.Shell import expand_vars

import inspect
import os
import re
import shutil
import signal
import time
import tempfile

from GangaCore.Utility.logging import getLogger

logger = getLogger()


class Interactive(IBackend):

    """Run jobs interactively on local host.

       Interactive job prints output directly on screen and takes the input from the keyboard.
       So it may be interupted with Ctrl-C
    """

    _schema = Schema(Version(1, 0), {
        "id": SimpleItem(defvalue=0, protected=1, copyable=0,
            doc="Process id"),
        "status": SimpleItem(defvalue="new", protected=1, copyable=0,
            doc="Backend status"),
        "exitcode": SimpleItem(defvalue=0, protected=1, copyable=0,
            doc="Process exit code"),
        "workdir": SimpleItem(defvalue="", protected=1, copyable=0,
            doc="Work directory"),
        "actualCE": SimpleItem(defvalue="", protected=1, copyable=0,
            doc="Name of machine where job is run")})

    _category = "backends"
    _name = 'Interactive'

    def __init__(self):
        super(Interactive, self).__init__()

    def _getIntFromOutfile(self, keyword="", outfileName=""):
        value = -999
        job = self.getJobObject()
        if keyword and outfileName and hasattr(job, "outputdir"):
            outfilePath = os.path.join(job.outputdir, outfileName)
            try:
                with open(outfilePath) as statfile:
                    statString = statfile.read()
                    testString = "".join(["^", keyword, " (?P<value>\\d*)"])
                    regexp = re.compile(testString, re.M)
                    match = regexp.search(statString)
                    if match:
                        value = int(match.group("value"))
            except IOError as err:
                logger.debug("IOError: %s" % str(err))
                pass
        return value

    def master_submit(self, rjobs, subjobconfigs, masterjobconfig, keep_going=False):
        """ Overload master_submit to avoid parallel submission with Interactive backend"""
        return IBackend.master_submit(self, rjobs, subjobconfigs, masterjobconfig, keep_going, False)

    def submit(self, jobconfig, master_input_sandbox):
        """Submit job to backend (i.e. run job interactively).

            Arguments other than self:
               subjobconfig         - Dictionary of subjob properties
               master_input_sandbox - Dictionary of properties of master job

            Return value: True always"""

        job = self.getJobObject()

        scriptpath = self.preparejob(jobconfig, master_input_sandbox)
        return self._submit(scriptpath, jobconfig.env)

    def resubmit(self):
        return self._submit(self.getJobObject().getInputWorkspace().getPath("__jobscript__"))

    def _submit(self, scriptpath, env=None):
        if env is None:
            env = expand_vars(os.environ)
        job = self.getJobObject()
        self.actualCE = util.hostname()
        logger.info('Starting job %s', job.getFQID('.'))

        try:
            job.updateStatus("submitted")
            self.status = "submitted"
            import subprocess
            subprocess.call(scriptpath)
            self.status = "completed"
        except KeyboardInterrupt:
            self.status = "killed"

        return True

    def kill(self):
        """Method for killing job running on backend

           No arguments other than self:

           Return value: True always"""

        job = self.getJobObject()

        if not self.id:
            time.sleep(0.2)
            self.id = self._getIntFromOutfile("PID:", "__id__")

        try:
            os.kill(self.id, signal.SIGKILL)
        except OSError as x:
            logger.warning("Problem killing process %d for job %d: %s" % (self.id, job.id, str(x)))

            self.status = "killed"
        self.remove_workdir()

        return True

    def remove_workdir(self):
        """Method for removing job's work directory

           No arguments other than self:

           Return value: None"""

        try:
            shutil.rmtree(self.workdir)
        except OSError as x:
            logger.warning("Problem removing workdir %s: %s", self.workdir, str(x))

            return None

    def preparejob(self, jobconfig, master_input_sandbox):
        """Method for preparing job script"""

        job = self.getJobObject()

        from GangaCore.GPIDev.Lib.File import File
        from GangaCore.Core.Sandbox.WNSandbox import PYTHON_DIR
        import GangaCore.Utility.files
        import inspect

        fileutils = File( inspect.getsourcefile(GangaCore.Utility.files), subdir=PYTHON_DIR )
        inputfiles = jobconfig.getSandboxFiles() + [ fileutils ]
        inbox = job.createPackedInputSandbox(inputfiles)

        inbox.extend(master_input_sandbox)
        inpDir = job.getInputWorkspace(create=True).getPath()
        outDir = job.getOutputWorkspace(create=True).getPath()
        workdir = tempfile.mkdtemp()
        self.workdir = workdir
        exeString = jobconfig.getExeString()
        argList = jobconfig.getArgStrings()
        argString = " ".join([" %s " % x for x in argList])

        outputSandboxPatterns = jobconfig.outputbox
        patternsToZip = []
        wnCodeForPostprocessing = ''
        wnCodeToDownloadInputFiles = ''

        if (len(job.outputfiles) > 0):

            from GangaCore.GPIDev.Lib.File.OutputFileManager import getOutputSandboxPatternsForInteractive, getWNCodeForOutputPostprocessing
            (outputSandboxPatterns,
                    patternsToZip) = getOutputSandboxPatternsForInteractive(job)

            wnCodeForPostprocessing = 'def printError(message):pass\ndef printInfo(message):pass' + \
                getWNCodeForOutputPostprocessing(job, '')
 
        all_inputfiles = [this_file for this_file in job.inputfiles]
        if job.master is not None: all_inputfiles.extend([this_file for this_file in job.master.inputfiles])

        wnCodeToDownloadInputFiles = ''

        if(len(all_inputfiles) > 0):

            from GangaCore.GPIDev.Lib.File.OutputFileManager import outputFilePostProcessingOnWN

            for inputFile in all_inputfiles:

                inputfileClassName = getName(inputFile)

                logger.debug("name: %s" % inputfileClassName)
                logger.debug("result: %s" % str(outputFilePostProcessingOnWN(job, inputfileClassName)))

                if outputFilePostProcessingOnWN(job, inputfileClassName):
                    inputFile.processWildcardMatches()
                    if inputFile.subfiles:
                        getfromFile = False
                        for subfile in inputFile.subfiles:
                            wnCodeToDownloadInputFiles += subfile.getWNScriptDownloadCommand('')
                        else:
                            getfromFile = True
                    else:
                        getFromFile = True

                    if getFromFile:
                        wnCodeToDownloadInputFiles += inputFile.getWNScriptDownloadCommand('')

        wnCodeToDownloadInputData = ''

        if job.inputdata and (len(job.inputdata) > 0):

            from GangaCore.GPIDev.Lib.File.OutputFileManager import getWNCodeForDownloadingInputFiles

            wnCodeToDownloadInputData = getWNCodeForDownloadingInputFiles(job, '')

        

        import inspect

        replace_dict = {
        '###CONSTRUCT_TIME###' : (time.strftime("%c")),
        '###WNSANDBOX_SOURCE###' : inspect.getsource(Sandbox.WNSandbox),
        '###GANGA_PYTHONPATH###' : getConfig("System")["GANGA_PYTHONPATH"],
        '###OUTPUTDIR###' : outDir,
        '###WORKDIR###' : workdir,
        '###IN_BOX###' : inbox,
        '###WN_INPUTFILES###' : wnCodeToDownloadInputFiles,
        '###WN_INPUTDATA###' : wnCodeToDownloadInputData,
        '###JOBCONFIG_ENV###' : jobconfig.env if jobconfig.env is not None else dict(),
        '###EXE_STRING###' : exeString,
        '###ARG_STRING###' : argString,
        '###WN_POSTPROCESSING###' : wnCodeForPostprocessing,
        '###PATTERNS_TO_ZIP###' : patternsToZip,
        '###OUTPUT_SANDBOX_PATTERNS###' : outputSandboxPatterns
        }

        script_location = os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))),
                'InteractiveScriptTemplate.py.template')

        from GangaCore.GPIDev.Lib.File import FileUtils
        commandString = FileUtils.loadScript(script_location, '')

        for k, v in replace_dict.items():
            commandString = commandString.replace(str(k), str(v))

        return job.getInputWorkspace().writefile(FileBuffer("__jobscript__", commandString), executable=1)

    @staticmethod
    def updateMonitoringInformation(jobs):

        for j in jobs:

            raw_backend = stripProxy(j.backend)

            if not j.backend.id:
                id = raw_backend._getIntFromOutfile("PID:", "__id__")
                if id > 0:
                    raw_backend.id = id
                    if ("submitted" == j.backend.status):
                        raw_backend.status = "running"

              # Check that the process is still alive
            if j.backend.id:
                try:
                    os.kill(j.backend.id, 0)
                except Exception as err:
                    logger.debug("Err: %s" % str(err))
                    raw_backend.status = "completed"

            if j.backend.status in ["completed", "failed", "killed"]:
                raw_backend.exitcode = raw_backend._getIntFromOutfile("EXITCODE:", "__jobstatus__")
               # Set job status to failed for non-zero exit code
                if j.backend.exitcode:
                    if j.backend.exitcode in [2, 9, 256]:
                        raw_backend.status = "killed"
                    else:
                        raw_backend.status = "failed"
                if (j.backend.status != j.status):
                    j.updateStatus(j.backend.status)

        return None




import re
from GangaCore.Core.exceptions import GangaException
from GangaCore.Utility.logging import getLogger
from GangaCore.GPIDev.Base.Proxy import addProxy, stripProxy, isType, getName

from GangaCore.GPIDev.Lib.GangaList.GangaList import GangaList, makeGangaListByRef

# config_scope is namespace used for evaluating simple objects (e.g. File, datetime, SharedDir)
from GangaCore.Utility.Config import config_scope

from GangaCore.Utility.Plugin import PluginManagerError, allPlugins

from GangaCore.GPIDev.Base.Objects import GangaObject, ObjectMetaclass
from GangaCore.GPIDev.Schema import Schema, Version
from GangaCore.GPIDev.Lib.GangaList.GangaList import makeGangaList

from .GangaRepository import SchemaVersionError

import xml.sax.saxutils
import copy
from io import StringIO

logger = getLogger()

_cached_eval_strings = {}

##########################################################################
# Ganga Project. http://cern.ch/ganga
#
# $Id: VStreamer.py,v 1.1.2.2 2009-07-14 09:20:22 ebke Exp $
##########################################################################

# dump object (job) to file f (or stdout) while ignoring the attribute
# 'ignore_subs'


class XMLFileError(GangaException):

    def __init__(self, excpt, message):
        GangaException.__init__(self, excpt, message)
        self.message = message
        self.excpt = excpt

    def __str__(self):
        if self.excpt:
            err = '(%s:%s)' % (type(self.excpt), self.excpt)
        else:
            err = ''
        return "XMLFileError: %s %s" % (self.message, err)

def _raw_to_file(j, fobj=None, ignore_subs=[]):
    sio = StringIO()
    vstreamer = VStreamer(out=sio, selection=ignore_subs)
    vstreamer.begin_root()
    j.accept(vstreamer)
    vstreamer.end_root()
    print(sio.getvalue(), file=fobj)

def to_file(j, fobj=None, ignore_subs=[]):
    #used to debug write problems - rcurrie
    #_raw_to_file(j, fobj, ignore_subs)
    #return
    _ignore_subs = [ignore_subs] if not isinstance(ignore_subs, list) else ignore_subs
    try:
        _raw_to_file(j, fobj, _ignore_subs)
    except Exception as err:
        logger.error("XML to-file error for file:\n%s" % (err))
        raise XMLFileError(err, "to-file error")

# Faster, but experimental version of to_file without accept()
# def to_file(j,f=None,ignore_subs=''):
#    sl = ["<root>"]
#    sl.extend(fastXML(j,'   ',ignore_subs=ignore_subs))
#    sl.append("</root>\n")
#    f.write("".join(sl))

# load object (job) from file f
# if len(errors) > 0 the object was not loaded correctly.
# Typical exceptions are:
# * SchemaVersionError (incompatible schema version)
# * PluginManagerError (necessary plugin not loaded)
# * IOError (problem on file reading)
# * AssertionError (corruption: multiple objects in <root>...</root>
# * Exception (probably corrupted data problem)

def _raw_from_file(f):
    # logger.debug('----------------------------')
    ###logger.debug('Parsing file: %s',f.name)
    xml_content = f.read()
    obj, errors = Loader().parse(xml_content)
    return obj, errors

def from_file(f):
    #return _raw_from_file(f)
    try:
        return _raw_from_file(f)
    except Exception as err:
        logger.error("XML from-file error for file:\n%s" % err)
        raise XMLFileError(err, "from-file error")

##########################################################################
# utilities


def escape(s):
    # s.replace('"', '\\"').replace("'", "\\'"))
    return xml.sax.saxutils.escape(s)


def unescape(s):
    return xml.sax.saxutils.unescape(s)

# An experimental, fast way to print a tree of Ganga Objects to file
# Unused at the moment


def fastXML(obj, indent='', ignore_subs=''):
    if hasattr(obj, "__iter__") and not hasattr(obj, "iteritems"):
        sl = ["\n", indent, "<sequence>", "\n"]
        for so in obj:
            sl.append(indent)
            sl.extend(fastXML(so, indent + ' ', ignore_subs))
        sl.append(indent)
        sl.append("</sequence>")
        return sl
    elif hasattr(obj, '_data'):
        v = obj._schema.version
        sl = ['\n', indent, '<class name="%s" version="%i.%i" category="%s">\n' % (getName(obj), v.major, v.minor, obj._category)]
        for attr_name in obj._schema.allItemNames():
            k = attr_name
            o = getattr(obj, k)
            if k != ignore_subs:
                try:
                    if not obj._schema[k]._meta["transient"]:
                        sl.append(indent)
                        sl.append('<attribute name="%s">' % k)
                        sl.extend(fastXML(o, indent + '  ', ignore_subs))
                        sl.append('</attribute>\n')
                except KeyError as err:
                    logger.debug("KeyError: %s" % err)
                    pass
        sl.append(indent)
        sl.append('</class>')
        return sl
    else:
        return ["<value>", escape(repr(obj)), "</value>"]

##########################################################################
# A visitor to print the object tree into XML.
#


class VStreamer(object):
    # Arguments:
    # out: file-like output stream where to print, default sys.stdout
    # selection: string specifying the name of properties which should not be printed
    # e.g. 'subjobs' - will not print subjobs

    def __init__(self, out=None, selection=[]):
        self.level = 0
        self.selection = selection
        if out is not None:
            self.out = out
        else:
            import sys
            self.out = sys.stdout

    def begin_root(self):
        print('<root>', file=self.out)

    def end_root(self):
        print('</root>', file=self.out)

    def indent(self):
        return ' ' * (self.level - 1) * 3

    def nodeBegin(self, node):
        self.level += 1
        s = node._schema
        print(self.indent(), '<class name="%s" version="%d.%d" category="%s">' % (s.name, s.version.major, s.version.minor, s.category), file=self.out)

    def nodeEnd(self, node):
        print(self.indent(), '</class>', file=self.out)
        self.level -= 1
        return

    def print_value(self, x):
        print('\n', self.indent(), '<value>%s</value>' % escape(repr(x)), file=self.out)

    def showAttribute(self, node, name):
        return (self.level > 1 or name not in self.selection) and not node._schema.getItem(name)['transient']

    def simpleAttribute(self, node, name, value, sequence):
        if self.showAttribute(node, name):
            self.level += 1
            print(self.indent(), end=' ', file=self.out)
            print('<attribute name="%s">' % name, end=' ', file=self.out)
            if sequence:
                self.level += 1
                print(file=self.out)
                print(self.indent(), '<sequence>', file=self.out)
                for v in value:
                    self.acceptOptional(v)
                print(self.indent(), '</sequence>', file=self.out)
                self.level -= 1
                print(self.indent(), '</attribute>', file=self.out)
            else:
                self.level += 1
                if isinstance(value, GangaObject):
                    print("", file=self.out)
                    self.acceptOptional(value)
                else:
                    self.print_value(value)
                self.level -= 1
                print(self.indent(), end=' ', file=self.out)
                print('</attribute>', file=self.out)
            self.level -= 1

    def sharedAttribute(self, node, name, value, sequence):
        self.simpleAttribute(node, name, value, sequence)

    def acceptOptional(self, s):
        self.level += 1
        if s is None:
            print(self.indent(), '<value>None</value>', file=self.out)
        else:
            if isType(s, str):
                print(self.indent(), '<value>%s</value>' % escape(repr(s)), file=self.out)
            elif hasattr(s, 'accept'):
                s.accept(self)
            elif isType(s, (list, tuple, GangaList)):
                print(self.indent(), '<sequence>', file=self.out)
                for sub_s in s:
                    self.acceptOptional(sub_s)
                print(self.indent(), '</sequence>', file=self.out)
            else:
                self.print_value(s)
        self.level -= 1

    def componentAttribute(self, node, name, subnode, sequence):
        if self.showAttribute(node, name):
            self.level += 1
            print(self.indent(), '<attribute name="%s">' % name, file=self.out)
            if sequence:
                self.level += 1
                print(self.indent(), '<sequence>', file=self.out)
                for s in subnode:
                    self.acceptOptional(s)
                print(self.indent(), '</sequence>', file=self.out)
                self.level -= 1
            else:
                self.acceptOptional(subnode)
            print(self.indent(), '</attribute>', file=self.out)
            self.level -= 1


##########################################################################
# XML Parser.

# Empty Ganga Object


class EmptyGangaObject(GangaObject):

    """Empty Ganga Object. Is used to construct incomplete jobs"""
    _schema = Schema(Version(0, 0), {})
    _name = "EmptyGangaObject"
    _category = "internal"
    _hidden = 1

    def __init__(self):
        super(EmptyGangaObject, self).__init__()


class Loader(object):

    """ Job object tree loader.
    """

    def __init__(self):
        self.stack = None  # we construct object tree using this stack
        # ignore nested XML elements in case of data errors at a higher level
        self.ignore_count = 0
        self.errors = []  # list of exception objects in case of data errors
        # buffer for <value> elements (evaled as python expressions)
        self.value_construct = None
        # buffer for building sequences (FIXME: what about nested sequences?)
        self.sequence_start = []

    def parse(self, s):
        """ Parse and load object from string s using internal XML parser (expat).
        """
        import xml.parsers.expat

        # 3 handler functions
        def start_element(name, attrs):
            #logger.debug('Start element: name=%s attrs=%s', name, attrs) #FIXME: for 2.4 use CurrentColumnNumber and CurrentLineNumber
            # if higher level element had error, ignore the corresponding part
            # of the XML tree as we go down
            if self.ignore_count:
                self.ignore_count += 1
                return

            # initialize object stack
            if name == 'root':
                assert self.stack is None, "duplicated <root> element"
                self.stack = []
                return

            assert not self.stack is None, "missing <root> element"

            # load a class, make empty object and push it as the current object
            # on the stack
            if name == 'class':
                try:
                    cls = allPlugins.find(attrs['category'], attrs['name'])
                except PluginManagerError as e:
                    self.errors.append(e)
                    #self.errors.append('Unknown class: %(name)s'%attrs)
                    obj = EmptyGangaObject()
                    # ignore all elemenents until the corresponding ending
                    # element (</class>) is reached
                    self.ignore_count = 1
                else:
                    version = Version(*[int(v) for v in attrs['version'].split('.')])
                    if not cls._schema.version.isCompatible(version):
                        attrs['currversion'] = '%s.%s' % (cls._schema.version.major, cls._schema.version.minor)
                        self.errors.append(SchemaVersionError('Incompatible schema of %(name)s, repository is %(version)s currently in use is %(currversion)s' % attrs))
                        obj = EmptyGangaObject()
                        # ignore all elemenents until the corresponding ending
                        # element (</class>) is reached
                        self.ignore_count = 1
                    else:
                        # Initialize and cache a c class instance to use as a classs factory
                        obj = cls.getNew()
                self.stack.append(obj)

            # push the attribute name on the stack
            if name == 'attribute':
                self.stack.append(attrs['name'])

            # start value_contruct mode and initialize the value buffer
            if name == 'value':
                self.value_construct = ''

            # save a marker where the sequence begins on the stack
            if name == 'sequence':
                self.sequence_start.append(len(self.stack))

        def end_element(name):
            #logger.debug('End element: name=%s', name)

            # if higher level element had error, ignore the corresponding part
            # of the XML tree as we go up
            if self.ignore_count:
                self.ignore_count -= 1
                return

            # when </attribute> is seen the current object, attribute name and
            # value should be on top of the stack
            if name == 'attribute':
                value = self.stack.pop()
                aname = self.stack.pop()
                obj = self.stack[-1]
                # update the object's attribute
                try:
                    obj.setSchemaAttribute(aname, value)
                except:
                    raise GangaException("ERROR in loading XML, failed to set attribute %s for class %s" % (aname, _getName(obj)))
                #logger.info("Setting: %s = %s" % (aname, value))

            # when </value> is seen the value_construct buffer (CDATA) should
            # be a python expression (e.g. quoted string)
            if name == 'value':
                try:
                    # unescape the special characters
                    s = unescape(self.value_construct)
                    s = re.sub(r'(\d)L(\})', r'\1\2', s)
                    if s not in _cached_eval_strings:
                        # This is ugly and classes which use this are bad, but this needs to be fixed in another PR
                        # TODO Make the scope of objects a lot better than whatever is in the config
                        # This is a dictionary constructed from eval-ing things in the Config. Why does should it do this?
                        # Anyway, lets save the result for speed
                        _cached_eval_strings[s] = eval(s, config_scope)
                    eval_str = _cached_eval_strings[s]
                    if not isinstance(eval_str, str):
                        val = copy.deepcopy(eval_str)
                    else:
                        val = eval_str
                    #logger.debug('evaled value: %s type=%s',repr(val),type(val))
                    self.stack.append(val)
                    self.value_construct = None
                except:
                    raise GangaException("ERROR in loading XML, failed to correctly parse attribute value: \'%s\'" % str(self.value_construct))

            # when </sequence> is seen we remove last items from stack (as indicated by sequence_start)
            # we make a GangaList from these items and put it on stack
            if name == 'sequence':
                pos = self.sequence_start.pop()
                try:
                    alist = makeGangaList(self.stack[pos:])
                except:
                    raise GangaException("ERROR in loading XML, failed to construct a sequence(list) properly")
                del self.stack[pos:]
                self.stack.append(alist)

            # when </class> is seen we finish initializing the new object
            # by setting remaining attributes to their default values
            # the object stays on the stack (will be removed by </attribute> or
            # is a root object)
            if name == 'class':
                obj = self.stack[-1]
                cls = obj.__class__
                if isinstance(cls, GangaObject):
                    for attr, item in cls._schema.allItems():
                        if attr not in obj._data:
                            if item.getProperties()['getter'] is None:
                                try:
                                    setattr(obj, attr, self._schema.getDefaultValue(attr))
                                except:
                                    raise GangaException("ERROR in loading XML, failed to set default attribute %s for class %s" % (attr, _getName(obj)))
                pass

        def char_data(data):
            # char_data may be called many times in one CDATA section so we need to build up
            # the full buffer for <value>CDATA</value> section incrementally
            if self.value_construct is not None:
                ###logger.debug('char_data: append=%s',data)
                # FIXME: decode data
                self.value_construct += data

        # start parsing using callbacks
        p = xml.parsers.expat.ParserCreate()

        p.StartElementHandler = start_element
        p.EndElementHandler = end_element
        p.CharacterDataHandler = char_data

        p.Parse(s)

        if len(self.stack) != 1:
            self.errors.append(AssertionError('multiple objects inside <root> element'))

        obj = self.stack[-1]

        #logger.info("obj.__dict__: %s" % obj.__dict__)

        # Raise Exception if object is incomplete
        for attr, item in obj._schema.allItems():
            if not hasattr(obj, attr):
                raise AssertionError("incomplete XML file")
        return obj, self.errors




################################################################################
# Ganga Project. http://cern.ch/ganga
#
################################################################################

from GangaCore.GPIDev.Base import GangaObject
from GangaCore.GPIDev.Adapters.IPostProcessor import PostProcessException
from GangaCore.GPIDev.Adapters.IChecker import IChecker, IFileChecker
from GangaCore.GPIDev.Base.Proxy import GPIProxyObject
from GangaCore.GPIDev.Schema import ComponentItem, FileItem, Schema, SimpleItem, Version
from GangaCore.Utility.Plugin import allPlugins
from GangaCore.Utility.logging import getLogger

import subprocess
import copy
import os
import string
import shutil

# Simon's post_status - communicates to processing DB
from . import post_status
import urllib.request, urllib.error, urllib.parse

logger = getLogger()

class FileCheckeR(IFileChecker):
    """
    Checks if string is in file.
    self.searchStrings are the search strings you would like to check for.
    self.files are the files you would like to check.
    self.failIfFound (default = True) decides whether to fail the job if the string is found. If you set this to false the job will fail if the string *isnt* found.
    self.filesMustExist toggles whether to fail the job if the specified file doesn't exist (default is True).
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['searchStrings'] = SimpleItem(defvalue = [], doc='String to search for')
    _schema.datadict['failIfFound'] = SimpleItem(True, doc='Toggle whether job fails if string is found or not found.')
    _category = 'postprocessor'
    _name = 'FileCheckeR'
    _exportmethods = ['check']


    def check(self,job):
        """
        Check that a string is in a file, takes the job object as input.
        """

        if not len(self.searchStrings):
            raise PostProcessException('No searchStrings specified, FileCheckeR will do nothing!')
        filepaths = self.findFiles(job)
        if not len(filepaths):
            raise PostProcessException('None of the files to check exist, FileCheckeR will do nothing!') 
        for filepath in filepaths:
            for searchString in self.searchStrings:
                grepoutput = subprocess.getoutput('grep "%s" %s' % (searchString,filepath))
                if len(grepoutput) and self.failIfFound is True:            
                    logger.info('The string %s has been found in file %s, FileCheckeR will fail job(%s)',searchString,filepath,job.fqid)
                    return self.failure
                if not len(grepoutput) and self.failIfFound is False:            
                    logger.info('The string %s has not been found in file %s, FileCheckeR will fail job(%s)',searchString,filepath,job.fqid)
                    return self.failure
        return self.result 


class ND280Kin_Checker(IFileChecker):
    """
    o Checks .log file (not impl. yet)
    o Checks if .kin file is present and non-zero
    o Moves output files to their destinations
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['prfx'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Path prefix to store output files')
    _schema.datadict['path'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Middle path to store output files')
    _category = 'postprocessor'
    _name = 'ND280Kin_Checker'
    _exportmethods = ['check']

    def check(self,job):
        """
        Checks if .kin file is present and non-zero
        """

        logger.info("Checking/moving outputs of run "+job.name)
        
        self.filesMustExist = False
        self.files = ['*.kin']
        filepaths = self.findFiles(job)
        if len(filepaths) != 1:
            logger.error('Something wrong with kin file(s) '+filepaths+'. CANNOT CONTINUE')
            self.move_output(job,ok=False)
            return False

        kinf = filepaths[0]
        if os.stat(kinf).st_size == 0:
            logger.error('Zero kin file '+kinf+'. CANNOT CONTINUE')
            self.move_output(job,ok=False)
            return False

        logger.info('OK')
        self.move_output(job,ok=True)
        return True    

        
    def move_output(self,job,ok=True):
        dest = os.path.join(self.prfx,self.path)
        jout = 'KIN_'+job.name
            
        task = {'*.kin':'kin','*.root':'kin','*.txt':'aux','*.conf':'aux','stdout':'jobOutput'}
        if not ok:
            task = {'*.kin':'errors','*.root':'errors','*.txt':'errors','*.conf':'errors','stdout':'errors'}
            
        for p in list(task.keys()):
            odir = os.path.join(dest,task[p])

            self.files = [p]
            filepaths = self.findFiles(job)
            for f in filepaths:
                if not os.path.isdir(odir):
                    os.makedirs(odir)

                shutil.move(f,odir)

                # rename stdout in odir
                if p == 'stdout':
                    shutil.move(os.path.join(odir,p),os.path.join(odir,jout))


class ND280RDP_Checker(IFileChecker):
    """
    o Checks .log file in a comprehensive way
    o Posts results to processing DB
    o Moves output files to there destinations
    """
    _schema = IFileChecker._schema.inherit_copy()
    _schema.datadict['prfx'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Path prefix to store output files')
    _schema.datadict['path'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Middle path to store output files')
    _schema.datadict['trig'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Trigger type, e.g. SPILL or COSMIC')
    _schema.datadict['site'] = SimpleItem(defvalue = None, typelist=['str','type(None)'], doc='Processing site, e.g. wg-bugaboo')
    _schema.datadict['post_status'] = SimpleItem(defvalue = None, typelist=['bool','type(None)'], doc='Post job information to processingstatus db')
    _category = 'postprocessor'
    _name = 'ND280RDP_Checker'
    _exportmethods = ['check']
    poststatus = True

    def __init__(self):
        super(ND280RDP_Checker,self).__init__()
        self.filesMustExist = False

        self.InStage = 0
        self.ReturnCode = -1
        self.EventsIn  = 0
        self.EventsOut = 0
        self.Time = 0
        self.STAGE = ''

        self.line = ''

        self.logf = ''
        self.RUN = ''
        self.SUBRUN = ''
        self.TRIGTYPE = None

        self.range = 1 # if to add 0000X000_0000X999 to paths (yes by default)
        
    def find(self,str):
        return self.line.find(str)>=0

    def reset_variables(self):
        self.InStage = 0
        self.ReturnCode = 1
        self.EventsIn  = 0
        self.EventsOut = 0
        self.Time = 0
        self.STAGE = ''

    def send_status(self):
        if not self.post_status:
            return 0

        logger.info('Result for %s %s %s %s is: %s, %s, %s, %s, %s' %  (self.RUN,self.SUBRUN,self.TRIGTYPE,self.STAGE,self.site,self.ReturnCode,self.Time,self.EventsIn,self.EventsOut))

        if self.range == 0: return # no remote status report for CosMC
        
        if not self.path:
            logger.error("No monitoring info sent because MONDIR is not defined")
            return

        rang = self.RUN[:5]+'000_'+self.RUN[:5]+'999'
        mondir = os.path.join(self.path,rang)
        job = { 'run':int(self.RUN), 'subrun':int(self.SUBRUN), 'trigtype':self.TRIGTYPE, 'stage':self.STAGE }
        attributes = {'site':self.site, 'result':self.ReturnCode,
                      'time':int(self.Time), 'read':self.EventsIn, 'written':self.EventsOut }

        try:
            logger.info("%s %s %s",str(mondir), str(job), str(attributes))
            post_status.record(mondir, job, attributes)
        except urllib.error.HTTPError as e:
            logger.error("Error: The HTTP request failed. (%s)" % (str(e), ))
            return 4

        return 0

    def check(self,job):

        IsMC = 0
        IsGenie = 0
        IsNeut = 0
        if self.trig == 'SPILL':
            self.TRIGTYPE = 'spill'
        elif self.trig == 'COSMIC':
            self.TRIGTYPE = 'cosmic'
        elif self.trig == 'MCP' or self.trig == 'MCSAND':
            self.TRIGTYPE = 'spill'
            IsMC = 1
        elif self.trig == 'MCCOS':
            self.TRIGTYPE = 'all'
            IsMC = 1
        else:
            #self.move_outs(job,ok=False)
            #return False
            raise PostProcessException("Unknown type of data: "+self.trig)

        if not self.site:
            #self.move_outs(job,ok=False)
            #return False
            raise PostProcessException('Site is not given')
            
        # finds .log file
        self.files = ['*.log']
        filepaths = self.findFiles(job)
        if len(filepaths) != 1:
            logger.error('Something wrong with logfile(s) '+filepaths+'. CANNOT CONTINUE')
            self.move_output(job,ok=False)
            return False
        self.logf = filepaths[0]

        filename = os.path.basename(self.logf)
        chunks = filename.split('_')
        chunks = chunks[3].split('-')
        self.RUN = chunks[0]
        self.SUBRUN = chunks[1]

        if not os.path.exists(self.logf):
            logger.error("Log file "+filename+" not found. Exit on error")
            self.move_output(job,ok=False)
            return False

        # Check generator if this is a beam MC
        if IsMC == 1:
            if filename.find('oa_nt_')>=0:
                logger.info("This is a NEUT MC log file")
                IsNeut = 1
            elif filename.find('oa_gn_')>=0:
                logger.info("This is a GENIE MC log file")
                IsGenie = 1

        if IsMC == 1:
            self.range = 0 # no range added to paths


        logger.info("Starting to scan file "+filename)
        logger.info("for run %s, subrun %s, type %s" % (self.RUN,self.SUBRUN,self.trig))
        
        inlogf = open(self.logf)
        for self.line in inlogf:
            self.line = self.line.strip('\n')
            if self.find('Midas File') and self.find('has been truncated'):
                logger.error('%s\n%s',self.line,"Midas file probably missing")
                self.ReturnCode = -1
                self.STAGE = "cali"
                self.send_status()
                self.InStage = 0
                break

            elif self.find('Starting job for neutMC.'):
                # neutMC logs are filled by Fluka, they are huge and seems useless
                logger.info(self.line+" The rest of log is ignored.")
                self.ReturnCode = 1
                self.STAGE = "neutMC"
                self.send_status()
                self.InStage = 0
                break
            
            elif self.find('Starting job for nd280MC'):
                self.InStage = 1
                self.STAGE = "nd280MC"
                logger.info(self.line)

            elif self.find('Starting job for elecSim'):
                self.InStage = 1
                self.STAGE = "elecSim"
                logger.info(self.line)

            elif self.find('Starting job for oaCosmicTrigger'):
                self.InStage = 1
                self.STAGE = "COSMICTRIG"
                logger.info(self.line)

            elif self.find('Starting job for oaCalib'):
                self.InStage = 1
                self.STAGE = "cali"
                logger.info(self.line)

            elif self.find('Starting job for oaRecon'):
                self.InStage = 1
                self.STAGE = "reco"
                logger.info(self.line)
 
            elif self.find('Starting job for oaAnalysis'):
                self.InStage = 1
                self.STAGE = "anal"
                logger.info(self.line)

            elif self.find('Starting job for '):
                logger.info(self.line)

            elif self.find('Found Command event_select '):
                logger.info(self.line)
                chunks = self.line.split()
                self.TRIGTYPE = chunks[5]

            elif self.InStage == 1:
                if   self.find('Segmentation fault'):
                    logger.error(self.line)
                    if self.line == '"oaCherryPicker-geo_v5mr.bat: line 7"':
                        logger.warning("This is an acceptable error - ignore it")
                    else:
                        self.ReturnCode = -2
                        self.InStage = 0
                        self.send_status()
                        break

                elif self.find('Disk quota exceeded'):
                    logger.error(self.line)
                    self.ReturnCode = -3
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find(' ERROR: No database for spillnum'):
                    logger.error(self.line)
                    self.ReturnCode = -8
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find(' No BSD data available'):
                    logger.error(self.line)
                    self.ReturnCode = -8
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('Disabling module '):
                    logger.info("IsMC = "+str(IsMC))
                    if IsMC == 1:
                        logger.info("After testing IsMC = "+str(IsMC))
                        if self.find('Disabling module GRooTrackerVtx') and IsGenie:
                            logger.error("Atest "+self.line)
                            self.ReturnCode = -4
                            self.InStage = 0
                            self.send_status()
                            break
                        elif self.find('Disabling module GRooTrackerVtx') and IsNeut:
                            logger.error("Btest "+self.line)
                            self.ReturnCode = -4
                            self.InStage = 0
                            self.send_status()
                            break
                        elif not self.find('RooTracker'):
                            logger.error("Ctest "+self.line)
                            self.ReturnCode = -4
                            self.InStage = 0
                            self.send_status()
                            break

                elif self.find('probably not closed, trying to recover'):
                    logger.error(self.line)
                    self.ReturnCode = -6
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('St9bad_alloc'):
                    logger.error(self.line)
                    self.ReturnCode = -7
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('No luck connecting to GSC MySQL server'):
                    logger.error(self.line)
                    self.ReturnCode = -9
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('EProductionException'):
                    logger.error(self.line)
                    self.ReturnCode = -11
                    self.InStage = 0
                    self.send_status()
                    break

                elif self.find('Total Events Read'):
                    chunks = self.line.split()
                    self.EventsIn = chunks[3]

                elif self.find('Total Events Written'):
                    chunks = self.line.split()
                    self.EventsOut = chunks[3]

                elif self.find('Number of events ='):
                    chunks = self.line.split()
                    if self.STAGE == 'nd280MC':
                        self.EventsOut = chunks[5]

                elif self.find('Total number of events processed in Analysis'):
                    chunks = self.line.split()
                    self.EventsOut = chunks[9]

                elif self.find('Job Completed Successfully'):
                    logger.info(self.line)
                    nextline = next(inlogf)
                    if nextline.find('Run time')>=0:
                        chunks = nextline.split()
                        Time = chunks[6]
                        chunks = Time.split('.')
                        self.Time = chunks[0]
                    self.ReturnCode = 1
                    if self.STAGE == 'COSMICTRIG':
                        logger.warning("Not sure what this stage is yet - no call to post_status")
                    else:
                        self.send_status()
                    self.reset_variables()
                    self.InStage = 0

        if self.InStage == 1:
            logger.error("The stage "+self.STAGE+" has not completed succesfully, Error is unknown")
            self.ReturnCode = 0
            self.send_status()
            
        inlogf.close()
        logger.info("Finished scanning the log file. Last check return code posted is "+str(self.ReturnCode))

        self.move_output(job)
        return self.ReturnCode == 1
    
    def move_output(self,job,ok=True):
        if ok:
            rang = self.RUN[:5]+'000_'+self.RUN[:5]+'999'
            dest = os.path.join(self.prfx,self.path)
            if self.range:
                dest = os.path.join(self.prfx,self.path,rang)
            jout = self.trig+'_'+self.RUN+'_'+self.SUBRUN
        else:
            dest = os.path.join(self.prfx,self.path)
            jout = 'stdout'
            
        task = {'*_g4mc_*.root':'g4mc','*_elmc_*.root':'elmc','*_cstr_*.root':'cstr',
                '*_numc_*.root':'numc','*_sand_*.root':'sand',
                '*_cali_*.root':'cali','*_reco_*.root':'reco','*_anal_*.root':'anal',
                '*.log':'logf','*catalogue.dat':'cata','stdout':'jobOutput'}
        if self.ReturnCode != 1 or not ok:
            task = {'*.log':'errors','*.root':'errors','*catalogue.dat':'errors','stdout':'errors'}
        for p in list(task.keys()):
            odir = os.path.join(dest,task[p])

            self.files = [p]
            filepaths = self.findFiles(job)
            for f in filepaths:
                if not os.path.isdir(odir):
                    os.makedirs(odir)

                shutil.move(f,odir)
                #os.symlink(f,os.path.join(odir,os.path.basename(f)))

                # rename stdout in odir
                if p == 'stdout':
                    shutil.move(os.path.join(odir,p),os.path.join(odir,jout))

                # "leave" symlink for _cstr_ or _numc_ files in job output directory
                # for possible transform chain
                if task[p] == 'cstr' or task[p] == 'numc':
                    os.symlink(os.path.join(odir,os.path.basename(f)),f)




# 05 Aug 2005 - KH : Added functions getSearchPath and getScriptPath

# 16 Aug 2005 - KH : Added method RuntimePackage.loadTemplates

# 30 Aug 2006 - KH : Modified function getSearchPath, to expand ~ and
#                    environment variables, and to allow paths to be
#                    specified relative to Ganga top directory

# 19 Oct 2006 - KH : Modified function getScriptPath, to expand ~ and
#                    environment variables

# 19 Oct 2006 - KH : Generalised function getSearchPath, allowing
#                    configuration parameter defining search path
#                    to be passed as arument

from GangaCore.Core.exceptions import PluginError

from GangaCore.Utility.util import importName

#from GangaCore.Utility.external.ordereddict import oDict
from GangaCore.Utility.external.OrderedDict import OrderedDict as oDict
allRuntimes = oDict()

import GangaCore.Utility.logging
logger = GangaCore.Utility.logging.getLogger(modulename=1)


def getScriptPath(name="", searchPath=""):
    """Determine path to a script

       Arguments:
          name       - Name of a script
          searchPath - String of colon-separated directory names,
                       defining locations to be searched for script

       If 'name' already gives the path to the script, then
       'searchPath' is ignored.

       Return value: String giving path to script (success)
                     or empty string (script not found)"""

    import os

    scriptPath = ""

    if name:
        fullName = os.path.expanduser(os.path.expandvars(name))
        if os.path.exists(fullName):
            scriptPath = fullName
        else:
            if searchPath:
                script = os.path.basename(fullName)
                pathList = searchPath.split(":")
                for directory in pathList:
                    if os.path.isdir(directory):
                        dirPath = os.path.abspath(directory)
                    if script in os.listdir(dirPath):
                        scriptPath = os.sep.join([dirPath, script])
                        break

    return scriptPath


def getSearchPath(configPar="SCRIPTS_PATH"):
    """Determine search path from configuration parameter

       Argument: 
          configPar : Name of configuration parameter defining search path

       Return value: Search path"""

    import os
    from GangaCore.Utility.Config import ConfigError, getConfig

    config = getConfig("Configuration")

    utilityDir = os.path.dirname(os.path.abspath(__file__))
    gangaRoot = os.path.dirname(os.path.dirname(utilityDir))

    pathString1 = ""
    if configPar:
        try:
            pathString1 = str(config[configPar])
        except ConfigError:
            logger.error("Option '%s' not defined in 'Configuration'" %
                         configPar)

    # always have . in the path in the first position!
    pathList1 = ['.'] + pathString1.split(":")
    pathList2 = []

    for path in pathList1:
        if ("." != path):
            path = os.path.expanduser(os.path.expandvars(path))
            if (0 != path.find("/")):
                path = os.path.join(gangaRoot, path)
        pathList2.append(path)

    pathString2 = ":".join(pathList2)
    return pathString2


class RuntimePackage(object):

    __slots__ = ('path', 'name', 'syspath', 'mod', 'modpath', 'config')

    def __init__(self, path):
        import os.path
        import sys
        import GangaCore.Utility.Config

        self.path = os.path.normpath(path.rstrip('/'))
        self.name = os.path.basename(self.path)
        self.syspath = os.path.dirname(self.path)
        self.mod = None
        self.modpath = ''

        showpath = self.syspath
        if not showpath:
            showpath = '<defaultpath>'
        logger.debug("initializing runtime: '%s' '%s'", self.name, showpath)

        if self.name in allRuntimes:
            if allRuntimes[self.name].path != self.path:
                logger.warning('possible clash: runtime "%s" already exists at path "%s"', self.name, allRuntimes[self.name].path)

        allRuntimes[self.path] = self

        if self.syspath:
            # FIXME: not sure if I really want to modify sys.path (side effects!!)
            # allow relative paths to GANGA_PYTHONPATH
            if not os.path.isabs(self.syspath):
                self.syspath = os.path.join(
                    GangaCore.Utility.Config.getConfig('System')['GANGA_PYTHONPATH'], self.syspath)
            sys.path.insert(0, self.syspath)

        # GangaCore.Utility.Config.getConfig('Runtime_'+self.name)
        self.config = {}

        try:
            self.mod = __import__(self.name)
            self.modpath = os.path.dirname(
                os.path.normpath(os.path.abspath(self.mod.__file__)))
            if self.syspath:
                if self.modpath.find(self.syspath) == -1:
                    logger.warning(
                        "runtime '%s' imported from '%s' but specified path is '%s'. You might be getting different code than expected!", self.name, self.modpath, self.syspath)
            else:
                logger.debug(
                    "runtime package %s imported from %s", self.name, self.modpath)

            # import the <PACKAGE>/PACKAGE.py module
            # @see Ganga/PACKAGE.py for description of this magic module
            # in this way we enforce any initialization of module is performed
            # (e.g PackageSetup.setPlatform() is called)
            __import__(self.name + ".PACKAGE")

        except ImportError as x:
            logger.warning("cannot import runtime package %s: %s", self.name, str(x))

    def standardSetup(self):
        """Perform any standard setup for the package"""
        g = importName(self.name, 'standardSetup')
        if g:
            return g()
        else:
            logger.debug("no standard setup defined for runtime package %s", self.name)
            return {}

    def loadPlugins(self):
        logger.debug("Loading Plugin: %s" % self.name)
        g = importName(self.name, 'loadPlugins')
        if g:
            g(self.config)
        else:
            logger.debug("no plugins defined for runtime package %s", self.name)
        logger.debug("Finished Plugin: %s" % self.name)

    def bootstrap(self, globals):
        try:
            import os.path

            # do not import names from BOOT file automatically, use
            # exportToGPI() function explicitly
            exec("import %s.BOOT" % self.name)
        except ImportError as x:
            logger.debug("problems with bootstrap of runtime package %s", self.name)
            logger.debug(x)
        except IOError as x:
            logger.debug("problems with bootstrap of runtime package %s", self.name)
            logger.debug(x)

    def loadNamedTemplates(self, globals, file_ext='tpl', pickle_files=False):
        try:
            import os
            from GangaCore.GPIDev.Lib.Registry.RegistryUtils import establishNamedTemplates
            template_registryname = 'templates%s' % self.name.strip('Ganga')
            template_pathname = os.path.join(self.modpath, 'templates')
            if os.path.isdir(template_pathname):
                establishNamedTemplates(template_registryname,
       template_pathname,
       "Registry for '%s' NamedTemplates" % self.name.strip(
           'Ganga'),
       file_ext=file_ext,
       pickle_files=pickle_files)
        except:
            logger.debug('failed to load named template registry')
            raise

    def loadTemplates(self, globals):
        try:
            import os.path
            if os.path.isfile(os.path.join(self.modpath, 'TEMPLATES.py')):
                exec(compile(open(os.path.join(self.modpath, 'TEMPLATES.py')).read(), os.path.join(self.modpath, 'TEMPLATES.py'), 'exec'), globals)
            else:
                logger.debug("Problems adding templates for runtime package %s",
                             self.name)
        except Exception as x:
            logger.debug\
                ("Problems adding templates for runtime package %s", self.name)
            logger.debug(x)

    def shutdown(self):
        g = importName(self.name, 'shutdown')
        if g:
            g()
        else:
            from GangaCore.Utility.logging import getLogger
            logger = getLogger(modulename=1)
            logger.debug("no shutdown procedure in runtime package %s", self.name)

    # this hook is called after the Ganga bootstrap procedure completed
    def postBootstrapHook(self):
        g = importName(self.name, 'postBootstrapHook')
        if g:
            g()
        else:
            logger.debug("no postBootstrapHook() in runtime package %s", self.name)


def initSetupRuntimePackages():
    """ Short wrapper function to init and setup Runtime Plugins """
    initRuntimePackages()

    # perform any setup of runtime packages
    logger.debug('Setting up Runtime Packages')
    for r in allRuntimes.values():
        r.standardSetup()


def initRuntimePackages():
    """
    initialize runtime packages, they are registered in allRuntimes
    dictionary automatically
    """
    import GangaCore.Utility.files
    from GangaCore.Utility.Config.Config import getConfig
    config = getConfig('Configuration')

    #if config['IgnoreRuntimeWarnings']:
    #    import warnings
    #    warnings.filterwarnings(action="ignore", category=RuntimeWarning)


    import inspect
    import os.path

    GangaRootPath = os.path.normpath(os.path.join(os.path.dirname(os.path.abspath(inspect.getfile(inspect.currentframe()))), '../..'))
    def transform(x):
        return os.path.normpath(GangaCore.Utility.files.expandfilename(os.path.join(GangaRootPath,x)))

    paths = list(map(transform, [_f for _f in [os.path.expandvars(os.path.expanduser(x)) for x in config['RUNTIME_PATH'].split(':')] if _f]))

    for path in paths:
        r = RuntimePackage(path)


def loadPlugins(environment):
    """
    Given a list of environments fully load the found Plugins into them exposing all of the relavent objects
    """
    from GangaCore.Utility.Runtime import allRuntimes
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    env_dict = environment.__dict__
    logger.debug("Loading: %s PLUGINS" % str(list(allRuntimes.keys())))
    for n, r in allRuntimes.items():
        logger.debug("Bootstrapping: %s" % n)
        try:
            r.bootstrap(env_dict)
        except Exception as err:
            logger.error('problems with bootstrapping %s -- ignored', n)
            logger.error('Reason: %s' % str(err))
            raise err
        try:
            r.loadNamedTemplates(env_dict, GangaCore.Utility.Config.getConfig('Configuration')['namedTemplates_ext'],
          GangaCore.Utility.Config.getConfig('Configuration')['namedTemplates_pickle'])
        except Exception as err:
            logger.error('problems with loading Named Templates for %s', n)
            logger.error('Reason: %s' % str(err))

    for n, r in allRuntimes.items():
        logger.debug("Loading: %s" % n)
        try:
            r.loadPlugins()
        except Exception as err:
            logger.error('problems with loading Plugin %s', n)
            logger.error('Reason: %s' % str(err))
            raise PluginError("Failed to load plugin: %s. Ganga will now shutdown to prevent job corruption." % n)

def autoPopulateGPI(my_interface=None):
    """
    Fully expose all plugins registered with the interface in a single line.
    By default only populate GPI, but also populate any other interface requested
    """
    if not my_interface:
        import GangaCore.GPI
        my_interface = GangaCore.GPI
    from GangaCore.Runtime.GPIexport import exportToInterface
    from GangaCore.Utility.Plugin import allPlugins
    # make all plugins visible in GPI
    for k in allPlugins.allCategories():
        for n in allPlugins.allClasses(k):
            cls = allPlugins.find(k, n)
            if not cls._declared_property('hidden'):
                if n != cls.__name__:
                    exportToInterface(my_interface, cls.__name__, cls, 'Classes')
                exportToInterface(my_interface, n, cls, 'Classes')

def setPluginDefaults(my_interface=None):
    """
    Set the plugin defaults for things like getting the defult plugin based upon class
    """
    from GangaCore.Utility.Plugin import allPlugins
    # set the default value for the plugins
    from GangaCore.Utility.Config import getConfig
    default_plugins_cfg = getConfig("Plugins")
    from GangaCore.Utility.logging import getLogger
    logger = getLogger()
    for opt in default_plugins_cfg:
        try:
            category, tag = opt.split('_')
        except ValueError as err:
            logger.warning("do not understand option %s in [Plugins]", opt)
            logger.debug('Reason: want %s' % str(err))
        else:
            if tag == 'default':
                try:
                    allPlugins.setDefault(category, default_plugins_cfg[opt])
                except GangaCore.Utility.Plugin.PluginManagerError as x:
                    logger.warning('cannot set the default plugin "%s": %s' % (opt, x))
            else:
                logger.warning("do not understand option %s in [Plugins]", opt)


    # set alias for default Batch plugin (it will not appear in the
    # configuration)

    batch_default_name = getConfig('Configuration').getEffectiveOption('Batch')
    try:
        batch_default = allPlugins.find('backends', batch_default_name)
    except Exception as x:
        from GangaCore.Utility.Config import ConfigError
        raise ConfigError('Check configuration. Unable to set default Batch backend alias (%s)' % str(x))
    else:
        allPlugins.add(batch_default, 'backends', 'Batch')
        from GangaCore.Runtime.GPIexport import exportToInterface
        if not my_interface:
            import GangaCore.GPI
            my_interface = GangaCore.GPI
        exportToInterface(my_interface, 'Batch', batch_default, 'Classes')






import os.path
import tempfile
from GangaCore.GPIDev.Lib.File import LocalFile
import GangaCore.Utility.logging
from GangaLHCb.Lib.LHCbDataset import LHCbDataset
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.Utility.files import expandfilename
logger = GangaCore.Utility.logging.getLogger()

## Due to a bug in Gaudi at some point we need this equivalence here: see #204
DataObjectDescriptorCollection = str

class PythonOptsCmakeParser(object):

    """ Parses job options file(s) w/ gaudirun.py to extract user's files
    Uses gaudirun.py to parse the job options file to allow for easy extraction
    of inputdata, outputdata and output files."""

    def __init__(self, optsfiles, app):
        self.optsfiles = optsfiles
        self.app = app
        self.opts_dict, self.opts_pkl_str = self._get_opts_dict_and_pkl_string()

    def _get_opts_dict_and_pkl_string(self):
        '''Parse the options using gaudirun.py and create a dictionary of the
        configuration and pickle the options. The app handler will make a copy
        of the .pkl file for each job.'''

        logger.info("Started parsing input Data file")

        tmp_pkl = tempfile.NamedTemporaryFile(suffix='.pkl', mode = "rb+")
        tmp_py = tempfile.NamedTemporaryFile(suffix='.py', mode = "rb+")
        py_opts = tempfile.NamedTemporaryFile(suffix='.py', mode = "w")
        py_opts.write(self._join_opts_files())
        py_opts.flush()
        gaudirun = './run gaudirun.py -n -v -o %s %s' % (tmp_py.name, py_opts.name)
        opts_str = ''
        err_msg = ''
        options = {}

        rc, stdout, m = self.app.execCmd(gaudirun)

    
        if stdout.decode().find('Gaudi.py') >= 0:
            msg = 'The version of gaudirun.py required for your application is not supported.'
            raise ValueError(None, msg)

        elif stdout.decode().find('no such option: -o') >= 0:
            gaudirun = './run gaudirun.py -n -v -p %s %s' % (tmp_pkl.name, py_opts.name)
            rc, stdout, m = self.app.execCmd(gaudirun)
            rc = 0

            if stdout and rc == 0:
                opts_str = stdout
                err_msg = 'Please check %s -v %s' % (cmdbase, py_opts.name)
                err_msg += ' returns valid python syntax'

        else:
            cmd = './run gaudirun.py -n -p %s %s' % (tmp_pkl.name, py_opts.name)
            rc, stdout, m = self.app.execCmd(cmd)
            if rc == 0 and stdout:
                opts_str = tmp_py.read()
                err_msg = 'Please check gaudirun.py -o file.py produces a valid python file.'

        if stdout and rc == 0:
            try:
                options = eval(opts_str)
            except Exception as err:
                logger.error('Cannot eval() the options file. Exception: %s', err)
                from traceback import print_exc
                logger.error(' ', print_exc())
                raise ApplicationConfigurationError(stdout + '###SPLIT###' + m)
            try:
                opts_pkl_string = tmp_pkl.read()
            except IOError as err:
                logger.error('Cannot read() the temporary pickle file: %s', tmp_pkl.name)
                raise err

        if not rc == 0:
            logger.debug('Failed to run: %s', gaudirun)
            raise ApplicationConfigurationError(stdout + '###SPLIT###' + m)

        tmp_pkl.close()
        py_opts.close()
        tmp_py.close()

        logger.info("Finished parsing input Data file")

        return (options, opts_pkl_string)

    def get_input_data(self):
        '''Collects the user specified input data that the job will process'''
        data = []
        try:
            opts_input = self.opts_dict['EventSelector']['Input']
            data = [f for f in opts_input]
        except KeyError as err:
            logger.error('No inputdata has been defined in the options file.')
            logger.error("%s" % str(err))

        from GangaCore.GPIDev.Base.Filters import allComponentFilters
        file_filter = allComponentFilters['gangafiles']

        all_files = []
        for d in data:
            p1 = d.find('DATAFILE=') + len('DATAFILE=')
            quote = d[p1]
            p2 = d.find(quote, p1 + 1)
            f = d[p1 + 1:p2]
            this_file = file_filter(f, None)
            if this_file is None:
                this_file = LocalFile(name=f)
            all_files.append(this_file)

        ds = LHCbDataset(files=all_files, fromRef=True)
        return ds

    def _join_opts_files(self):
        '''Create a single options file from all supplied options.'''
        joined_py_opts = ''
        for name in self.optsfiles:
            with open(expandfilename(name), 'r') as this_file:
                joined_py_opts += this_file.read()
        return joined_py_opts




from GangaCore.GPIDev.Schema import *
from GangaCore.GPIDev.Lib.Tasks.common import *
from GangaCore.GPIDev.Lib.Tasks.IUnit import IUnit
from GangaCore.GPIDev.Lib.Job.Job import JobError
from GangaCore.GPIDev.Lib.Registry.JobRegistry import JobRegistrySlice, JobRegistrySliceProxy
from GangaCore.Core.exceptions import ApplicationConfigurationError
from GangaCore.GPIDev.Base.Proxy import addProxy, stripProxy
from GangaCore.Utility.logging import getLogger
import GangaCore.GPI as GPI

import os
from copy import deepcopy

logger = getLogger()

class ND280Unit(IUnit):
   _schema = Schema(Version(1,0), dict(IUnit._schema.datadict.items()))

   _category = 'units'
   _name = 'ND280Unit'
   _exportmethods = IUnit._exportmethods + [ ]

   def __init__(self):
      super(ND280Unit, self).__init__()
      
   def createNewJob(self):
      """Create any jobs required for this unit"""      
      j = GPI.Job()
      j._impl.backend = self._getParent().backend.clone()
      j._impl.application = self._getParent().application.clone()
      if not self.inputdata is None:
        j.inputdata = self.inputdata.clone()

      trf = self._getParent()
      task = trf._getParent()

      # copy across the outputfiles
      for f in trf.outputfiles:
         j.outputfiles += [f.clone()]

      j.inputsandbox = trf.inputsandbox

      # Sort out the splitter
      if trf.splitter:
         j.splitter = trf.splitter.clone()

      # Postprocessors
      for pp in trf.postprocessors:
         j.postprocessors.append(deepcopy(pp))
         
      return j

   def checkMajorResubmit(self, job):
      """Check if a failed job shold be 'rebrokered' (new job created) rather than just resubmitted"""
      return False

   def majorResubmit(self, job):
      """Do any bookkeeping before a major resubmit is done"""
      super(ND280Unit,self).majorResubmit(job)

   def reset(self):
      """Reset the unit completely"""
      super(ND280Unit,self).reset()

   def updateStatus(self, status):
      """Update status hook"""
      super(ND280Unit,self).updateStatus(status)

   def checkForSubmission(self):
      """Additional checks for unit submission"""

      # call the base class
      if not super(ND280Unit,self).checkForSubmission():
         return False

      return True

   def copyOutput(self):
      """Copy the output data to local storage"""

      job = GPI.jobs(self.active_job_ids[0])
      
      if self.copy_output._name != "TaskLocalCopy" or job.outputdata._impl._name != "DQ2OutputDataset":
         logger.error("Cannot transfer from DS type '%s' to '%s'. Please contact plugin developer." % (job.outputdata._name, self.copy_output._name))
         return False

      # check which fies still need downloading
      to_download = []
      for f in job.outputfiles:
         
         # check for REs
         if self.copy_output.isValid( os.path.join(f.localDir, f.namePattern)) and not self.copy_output.isDownloaded( os.path.join(f.localDir, f.namePattern)):
            to_download.append(f)

      # is everything downloaded?
      if len(to_download) == 0:
         return True

      # nope, so pick the requested number and off we go
      for f in to_download:
         f.get()

      return False
